/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: ";
#endif

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* AGFA camera maker-specific routines                                */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* This information is from a page by Johannes Tschebisch at:         */
/* http://www.jojotsch.de/downloads/jojothumb/beispiele/              */
/*                             html_exif/bilder/agfa-c150.html        */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "maker.h"

/* Find the identifying number assigned to known Agfa camera          */
/* models. This number is used to dispatch print and interpret        */
/* routines approopriate to the current image.                        */

int
agfa_model_number(char *model,char *software)
{
    int number = GENERIC;

    if(model)
    {
        switch(*model)
        {
            case 'S':
                if(strcmp(model,"SR841") == 0)
                    number = AGFA_SR841;
                break;
            default:
                number = GENERIC;
                break;
        }
    }
    return(number);
}

/* Dispatch a print routine based upon model                          */

void
print_agfa_makervalue(struct ifd_entry *entry_ptr,int make, int model,
                    char *prefix)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            case AGFA_SR841:
            case GENERIC:
            default:
                print_agfa1_makervalue(entry_ptr,make,model,PREFIX);
                agfa1_interpret_value(entry_ptr);
                break;
        }
    }
}

/* Model-specific print routine for Agfa cameras. This routine is     */
/* responsible for picking off any direct entry tags which are        */
/* peculiar and will not be handled properly by print_value()         */
/* (usually UNDEFINED values which fit in the 4-byte entry value). If */
/* there are no such entries, this function simply calls              */
/* print_value().                                                     */

void
print_agfa1_makervalue(struct ifd_entry *entry_ptr,int make, int model,
                    char *prefix)
{
    if(entry_ptr && (PRINT_VALUE))
    {
        switch(entry_ptr->tag) 
        {
            default:
                print_value(entry_ptr,PREFIX);
                break;
        }
    }
}

/* Dispatch a routine to decode and print offset values for Agfa      */
/* cameras.                                                           */
void
print_agfa_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            default:            /* default to the new version             */
                agfa1_offset_makervalue(inptr,byteorder,entry_ptr,
                                            offset_ref,prefix,indent,
                                            addrwidth,make,model,at_offset);
                agfa1_interpret_offset_makervalue(inptr,byteorder,entry_ptr,
                                            offset_ref);
                break;
        }
    }
}

/* Model-specific routine to print UNDEFINED values found at offsets  */
/* in OLYMPUS makernotes. One of these may be supplied for each model */
/* with unique UNDEFINED tags.                                        */

void
agfa1_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{
    unsigned long value_offset;
    unsigned long dumplength;
    unsigned short marker;
    char *nameoftag;

    if(entry_ptr)
    {
        nameoftag = maker_tagname(entry_ptr->tag,make,model); 
        value_offset = offset_ref + entry_ptr->value;
        switch(entry_ptr->tag)
        {
            case 0x0209:    /* CameraId                               */
                if(at_offset)
                {
                    print_tag_address(ENTRY,value_offset,indent,prefix,addrwidth);
                    print_makertagid(entry_ptr,19," = ",SUBINDENT,make,model);
                }
                print_ascii(inptr,entry_ptr->count,value_offset);
                break;
            case 0x0100: /* JPEG Thumbnail in TIFF MakerNote          */
                if(at_offset)
                {
                    print_tag_address(VALUE_AT_OFFSET,value_offset,indent,prefix,addrwidth);
                    extraindent(SUBINDENT);
                    printf("Start of JPEG Thumbnail");
                }
                if(!(PRINT_OFFSET))
                {
                    if(PRINT_BOTH_VALUE)
                        printf("@%#lx=%lu",value_offset,value_offset);
                    else if(PRINT_HEX_VALUE)
                        printf("@%#lx",value_offset);
                    else if(PRINT_DEC_VALUE)
                        printf("@%lu",value_offset);
                }
                printf(" length %ld",entry_ptr->count);
                marker = read_imagemarker(inptr,value_offset);
                if(PRINT_SEGMENT)
                {
                    putchar('\n');
                    (void)process_jpeg_segments(inptr,value_offset,
                                                marker,entry_ptr->count,
                                                indent+(SUBINDENT * 2),addrwidth);
                    print_tag_address(SEGMENT,value_offset + entry_ptr->count - 1,
                                                indent,"-",addrwidth);
                    extraindent(SUBINDENT);
                    printf("End of JPEG Thumbnail");
                }
                /* make certain we're at the end unless 'inptr' bad   */
                fseek(inptr,value_offset + entry_ptr->count,0);
                break;
            case 0x0f00: /* Data                                      */
                if(at_offset)
                {
                    print_tag_address(ENTRY,value_offset,indent,prefix,addrwidth);
                    print_makertagid(entry_ptr,19," : ",SUBINDENT,make,model);
                }
                else if(PRINT_OFFSET)
                        printf(" = ");
                printf("UNDEFINED, count %lu", entry_ptr->count);
                if(Max_undefined == 0)
                    printred(" (not dumped, use -U)");
                else
                {
                    if((Max_undefined == DUMPALL)
                        || (Max_undefined > entry_ptr->count))
                            dumplength = entry_ptr->count;
                    else 
                        dumplength = Max_undefined;
                    putchar('\n');
                    hexdump(inptr,value_offset,entry_ptr->count,dumplength,12,
                                indent,SUBINDENT,addrwidth);
                }
                /* make certain we're at the end unless 'inptr' bad   */
                fseek(inptr,value_offset + entry_ptr->count,0);
                break;
            default:
                print_generic_offset_makervalue(inptr,byteorder,entry_ptr,
                                        offset_ref,prefix,indent,
                                        addrwidth,make,model,at_offset);
                break;
        }
    }
}

/* Agfa-specific tagnames for makernotes.                             */

char *
maker_agfa_tagname(unsigned short tag,int model)
{
    char *tagname = (char *)0;

    /* Check specific models first                                    */
    switch(model)
    {
        /* Add special-case models here                               */
        default:
            break;
    }

    /* If no model-specific tag is found, use a "generic" Agfa tag    */
    if(tagname == NULL)
    {
        switch(tag)
        {
            case 0x0100: tagname = "JPEG_Thumbnail"; break;
            case 0x0200: tagname = "SpecialMode"; break;
            case 0x0201: tagname = "Quality"; break;
            case 0x0202: tagname = "Macro"; break;
            case 0x0204: tagname = "DigitalZoom"; break;
            case 0x0207: tagname = "SoftwareRelease"; break;
            case 0x0208: tagname = "PictureInfo"; break;
            case 0x0209: tagname = "CameraID"; break;
            case 0x020b: tagname = "ImageWidth"; break; /* ? */
            case 0x020c: tagname = "ImageHeight"; break; /* ? */
            case 0x0f00: tagname = "Data"; break;
            default: break;
        }
    }
    return(tagname);
}


void
agfa1_interpret_value(struct ifd_entry *entry_ptr)
{
    switch(entry_ptr->tag)
    {
        case 0x0201:    /* Compression Mode */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 1: printf("SQ"); break;
                case 2: printf("HQ"); break;
                case 3: printf("SHQ"); break;
                case 33: printf("not compressed"); break; /* TIFF  */
                case 34:    /* not sure about this one ###%%%         */
                            printf("not compressed");
                            break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0202:    /* Macro */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Normal"); break;
                case 1: printf("Macro"); break;
                case 2: printf("View"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        default:
            break;
    }
}

void
agfa1_interpret_offset_makervalue(FILE *inptr,unsigned short byteorder,
                    struct ifd_entry *entry_ptr,unsigned long offset_ref)
{
    unsigned long offset,value;

    if(entry_ptr)
    {
        offset = entry_ptr->value + offset_ref;
        switch(entry_ptr->tag)
        {
            case 0x0200:    /* SpecialMode                            */
                value = read_ulong(inptr,byteorder,offset);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Normal"); break;
                    case 1: printf("Unkown"); break;
                    case 2: printf("Fast"); break;
                    case 3: printf("Panorama,");
                            value = read_ulong(inptr,byteorder,HERE);
                            printf("#%lu,",value);
                            value = read_ulong(inptr,byteorder,HERE);
                            switch(value)
                            {
                                case 1: printf(" Left to Right"); break;
                                case 2: printf(" Right to Left"); break;
                                case 3: printf(" Bottom to Top"); break;
                                case 4: printf(" Top to Bottom"); break;
                                default: printred(" undefined"); break;
                            }
                            break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            default:
                break;
        }
    }
}
