/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: maker_epson.c,v 1.13 2003/04/26 14:15:17 alex Exp $";
#endif

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Epson camera maker-specific routines                               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* This is currently a near-duplicate of the Olympus, based upon the  */
/* similarity of tag numbers, the appearance of sample images taken   */
/* from the Web, and reports that some Olympus, Epson, Minolta and    */
/* Sanyo models use similar firmware.                                 */
/* The "photopc" site at soucrceforge provides some insight:          */
/*    http://photopc.sourceforge.net/cameras.html                     */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "maker.h"


/* Find the identifying number assigned to known Epson camera models. */
/* This number is used to dispatch appropriate print and interpret    */
/* routines approopriate to the current image.                        */

int
epson_model_number(char *model,char *software)
{
    int number = GENERIC;

    if(model)
    {
        switch(*model)
        {
            case 'P':
                if(strcmp(model,"PhotoPC 3000Z") == 0)
                    number = EPSON_3000Z;
                else if(strcmp(model,"PhotoPC 3100Z") == 0)
                    number = EPSON_3100Z;
                break;
            default:
                number = GENERIC;
                break;
        }
    }
    return(number);
}


/* Dispatch a print routine for direct IFD values in Epson            */
/* MakerNotes, based upon model                                       */

void
print_epson_makervalue(struct ifd_entry *entry_ptr,int make, int model,
                    char *prefix)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            case EPSON_3000Z:
            case EPSON_3100Z:
            case GENERIC:
            default:
                print_epson1_makervalue(entry_ptr,make,model,prefix);
                epson1_interpret_value(entry_ptr);
                break;
        }
    }
}

/* Model-specific print routine for Epson cameras. This routine is      */
/* responsible for picking off any direct entry tags which are        */
/* peculiar and will not be handled properly by print_value()         */
/* (usually UNDEFINED values which fit in the 4-byte entry value). If */
/* there are no such entries, this function simply calls              */
/* print_value().                                                     */

void
print_epson1_makervalue(struct ifd_entry *entry_ptr,int make, int model,
                    char *prefix)
{
    if(entry_ptr && (PRINT_VALUE))
    {
        switch(entry_ptr->tag) 
        {
            default:
            print_value(entry_ptr,PREFIX);
            break;
        }
    }
}

/* Dispatch a routine to decode and print offset values in MakerNotes */
/* from Epson cameras.                                                */

void
print_epson_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            default:
                epson1_offset_makervalue(inptr,byteorder,entry_ptr,
                                            offset_ref,prefix,indent,
                                            addrwidth,make,model,at_offset);
                epson1_interpret_offset_makervalue(inptr,byteorder,
                                            entry_ptr,offset_ref);
                break;
        }
    }
}

/* Model-specific routine to print UNDEFINED values found at offsets  */
/* in epson makernotes.                                               */

void
epson1_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{
    unsigned long value_offset;
    unsigned long dumplength;
    unsigned short marker;
    char *nameoftag;

    if(entry_ptr)
    {
        nameoftag = maker_tagname(entry_ptr->tag,make,model); 
        value_offset = offset_ref + entry_ptr->value;
        switch(entry_ptr->tag)
        {
            case 0x020a:    /* MakerNote version of UserComment   */
                if(at_offset)
                {
                    print_tag_address(ENTRY,value_offset,indent,prefix,addrwidth);
                    print_makertagid(entry_ptr,19," = ",SUBINDENT,make,model);
                }
                else if(PRINT_OFFSET)
                        printf(" = ");
                print_user_comment(inptr,entry_ptr->count,value_offset);
                break;
            case 0x0209:    /* CameraId                           */
                if(at_offset)
                {
                    print_tag_address(ENTRY,value_offset,indent,prefix,addrwidth);
                    print_makertagid(entry_ptr,19," = ",SUBINDENT,make,model);
                }
                else if(PRINT_OFFSET)
                        printf(" = ");
                print_ascii(inptr,entry_ptr->count,value_offset);
                break;
            case 0x0100: /* JPEG Thumbnail in TIFF MakerNote      */
                if(at_offset)
                {
                    print_tag_address(VALUE_AT_OFFSET,value_offset,indent,prefix,addrwidth);
                    extraindent(SUBINDENT);
                    printf("Start of JPEG Thumbnail");
                }
                if(!(PRINT_OFFSET))
                {
                    if(PRINT_BOTH_VALUE)
                        printf("@%#lx=%lu",value_offset,value_offset);
                    else if(PRINT_HEX_VALUE)
                        printf("@%#lx",value_offset);
                    else if(PRINT_DEC_VALUE)
                        printf("@%lu",value_offset);
                }
                printf(" length %ld",entry_ptr->count);
                marker = read_imagemarker(inptr,value_offset);
                if(PRINT_SEGMENT)
                {
                    putchar('\n');
                    (void)process_jpeg_segments(inptr,value_offset,
                                                marker,entry_ptr->count,
                                                indent+(SUBINDENT * 2),addrwidth);
                    print_tag_address(SEGMENT,value_offset + entry_ptr->count - 1,
                                                indent,"-",addrwidth);
                    extraindent(SUBINDENT);
                    printf("End of JPEG Thumbnail");
                }
                /* make certain we're at the end unless 'inptr' bad   */
                fseek(inptr,value_offset + entry_ptr->count,0);
                break;
            case 0x0e00: /* PrintIM (Epson Print Image matching)  */
            case 0x0f00:    /* Data                               */
                if(at_offset)
                {
                    print_tag_address(ENTRY,value_offset,indent,prefix,addrwidth);
                    print_makertagid(entry_ptr,19," : ",SUBINDENT,make,model);
                }
                else if(PRINT_OFFSET)
                        printf(" = ");
                printf("UNDEFINED, count %lu", entry_ptr->count);
                if(Max_undefined == 0)
                    printred(" (not dumped, use -U)");
                else
                {
                    if((Max_undefined == DUMPALL)
                        || (Max_undefined > entry_ptr->count))
                            dumplength = entry_ptr->count;
                    else 
                        dumplength = Max_undefined;
                    putchar('\n');
                    hexdump(inptr,value_offset,entry_ptr->count,dumplength,12,
                                indent,SUBINDENT,addrwidth);
                }
                /* make certain we're at the end                  */
                fseek(inptr,value_offset + entry_ptr->count,0);
                break;
            default:
                print_generic_offset_makervalue(inptr,byteorder,entry_ptr,
                                        offset_ref,prefix,indent,
                                        addrwidth,make,model,at_offset);
                break;
        }
    }
}

/* Epson-specific tagnames for makernotes.                            */

char *
maker_epson_tagname(unsigned short tag,int model)
{
    char *tagname = (char *)0;

    /* Check specific models first                                    */
    switch(model)
    {
        /* Add special-case models here                               */
        default:
            break;
    }

    /* The only thing I know for sure about the Epson MakerNote is    */
    /* that the TIFF notes have a JPEG thumbnail. I suspect the tags  */
    /* may be similar to the Olympus (at least, they are for images I */
    /* have seen).                                                    */
    if(tagname == NULL)
    {
        switch(tag)
        {
            case 0x0100: tagname = "JPEG_Thumbnail"; break;
            default: break;
        }
    }
    
    /* These are the Olympus tags; try them ###%%% ???                */
    if(tagname == NULL)
    {
        switch(tag)
        {
            case 0x0200: tagname = "SpecialMode"; break;
            case 0x0201: tagname = "CompressionMode"; break;
            case 0x0202: tagname = "Macro"; break;
            case 0x0204: tagname = "DigitalZoom"; break;
            case 0x0207: tagname = "SoftwareRelease"; break;
            case 0x0208: tagname = "PictureInfo"; break;
            case 0x0209: tagname = "CameraID"; break;
            case 0x020a: tagname = "MakerUserComment"; break;
            case 0x020b: tagname = "ImageWidth"; break;
            case 0x020c: tagname = "ImageHeight"; break;
            case 0x0e00: tagname = "PrintIMData"; break;
            case 0x0f00: tagname = "Data"; break;
            default: break;
        }
    }
    return(tagname);
}


/* This arbitrarily assumes that the Epsons use firmware similar to   */
/* that of the Olympus. I have seen no information on the web from    */
/* Epson owners who have decoded notes from their cameras. The        */
/* following seems to be correct based upon Epson PhotoPC 3000Z       */
/* images downloaded from "dpreview" and "imaging resources".         */

void
epson1_interpret_value(struct ifd_entry *entry_ptr)
{
    switch(entry_ptr->tag)
    {
        case 0x0201:    /* Compression Mode */
            print_startvalue();
            /* 'Quality' is a mix of this and image size.             */
            /* Interestingly, 'HyPict' (2544x1904) interpolated       */
            /* images are marked '34' (same as uncompressed 'tiff')   */
            /* but are in JPEG format with CompressedBitsPerPixel=1.  */
            /* The value '33' is used by Olympus cameras, which       */
            /* appear to use similar software Of course, the Minolta  */
            /* F100 looks a lot like this too, but uses '4' for       */
            /* uncompressed.                                          */
            switch(entry_ptr->value)
            {
                case 1: printf("SQ"); break;
                case 2: printf("HQ"); break;
                case 3: printf("SHQ"); break;
                case 33: printf("not compressed"); break;
                case 34: printf("not compressed"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0202:    /* Macro */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Normal"); break;
                case 1: printf("Macro"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        default:
            break;
    }
}

/* Report the "meaning" of tag values found at offsets in an Epson    */
/* MakerNote IFD entry (not at an offset).                            */

void
epson1_interpret_offset_makervalue(FILE *inptr,unsigned short byteorder,
                    struct ifd_entry *entry_ptr,unsigned long offset_ref)
{
    unsigned long offset,value;

    if(entry_ptr)
    {
        offset = entry_ptr->value + offset_ref;
        switch(entry_ptr->tag)
        {
            case 0x0200:    /* SpecialMode                            */
                value = read_ulong(inptr,byteorder,offset);
                print_startvalue();
                /* Don't know what the last 2 values might mean for   */
                /* "normal" or "fast".                                */
                switch(value)
                {
                    case 0: printf("Normal"); break;
                    case 1: printf("Unkown"); break;
                    case 2: printf("Fast"); break;
                    case 3: printf("Panorama,");
                            value = read_ulong(inptr,byteorder,HERE);
                            printf("#%lu,",value);
                            value = read_ulong(inptr,byteorder,HERE);
                            switch(value)
                            {
                                case 1: printf(" Left to Right"); break;
                                case 2: printf(" Right to Left"); break;
                                case 3: printf(" Bottom to Top"); break;
                                case 4: printf(" Top to Bottom"); break;
                                default: printred(" undefined"); break;
                            }
                            break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            default:
                break;
        }
    }
}
