/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: maker_generic.c,v 1.22 2003/08/02 02:10:35 alex Exp $";
#endif

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* "Generic" routines for makernotes which use TIFF IFD structure     */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "maker.h"

/* Read a makernote at the offset given in the passed maker_entry.    */

/* Checks for data with a TIFF IFD structure at the makernote offset, */
/* possibly offset by a maker name string and/or a new TIFF header    */
/* (e.g. some later Nikons). Notes which pass the test are scanned    */
/* and printed. If the note is in IFD format but the maker is         */
/* unknown, a "bare bones" entry will be printed, with tag numbers,   */
/* type, etc. If the maker is known, tag names, undefined values and  */
/* "interpretations" of the data may be printed. MakerNotes with      */
/* unknown/non-ifd structures could be dispatched to specialized      */
/* routines, but at this time I do not have information on any of     */
/* those. Unknown notes (or notes with invalid IFD entries) can be    */
/* dumped in hex/ascii by user option (-M). 'Undefined" data may be   */
/* similarly dumped by the -U option, if maker-specific routines      */
/* don't handle it.                                                   */

/* The return is the offset of the end of the note; if the note       */
/* cannot be read, return is 0L.                                      */

unsigned long
process_makernote(FILE *inptr,unsigned short byteorder,
                    struct ifd_entry *maker_entry,
                    unsigned long fileoffset_base,
                    unsigned long max_offset,
                    int indent,int addrwidth)
{
    struct ifd_entry *entry_ptr;
    struct maker_scheme *maker_scheme;
    unsigned long maker_offset = 0UL;
    unsigned long maker_count = 0UL;
    unsigned long max_value_offset = 0UL;
    unsigned long start_offset = 0UL;
    unsigned long dumplength = 0UL;
    unsigned long end_of_note = 0UL;
    unsigned long next_ifd_offset,current_offset;
    unsigned long start_entry_offset;
    unsigned long entry_offset,value_offset;
    char *prefix;
    int entry_num,num_entries,make,model;
    int invalid = 0;
    int value_is_offset = 0;
    int note_relative_offset_scheme = 0;
    int use_second_pass = 0;

    if(inptr && maker_entry)
    {
        STARTCOLOR(MAKER_COLOR);
        /* Save these; they're in static storage in the maker_entry   */
        /* and will be overwritten when entry reading starts.         */
        maker_offset = maker_entry->value;
        maker_count = maker_entry->count;
        if(maker_count > 4UL)
            end_of_note = maker_offset + maker_count + fileoffset_base;
        else if(maker_count == 4UL)
        {
            /* special case, for certain Casio MakerNotes which use   */
            /* 4-byte UNDEFINED as though it were a LONG offset       */
            if(maker_entry->value_type == UNDEFINED)
                maker_offset = to_ulong((unsigned char *)&maker_entry->value,byteorder);
            /* This pretty much forces maker_scheme to PLAIN          */
            maker_count=0L;
        }
        else
            goto cant_dumpit;

        /* Get make and model from the global names saved from IFD0   */
        make = maker_number(Make_name);
        model = model_number(make,Model_name,Software_name);

        /* start_offset is where the note begins, relative to the     */
        /* start of the file.                                         */
        start_offset = start_entry_offset = maker_offset + fileoffset_base;

        /* Mark the beginning offset of the note info. At least one   */
        /* maker writes the makernote past the end of parent IFD so   */
        /* that it is not actually contained within the parent (EXIF) */
        /* IFD. The following attempts to mark that situation by      */
        /* using a '^' to mark the start address, so that it stands   */
        /* out a little. If "max_offset" is non-zero, it is presumed  */
        /* to be the highest offset of the parent, ususally taken     */
        /* from a 'next_ifd_offset' field.                            */
        if(max_offset && (start_offset > max_offset))
            prefix = "^";
        else
            prefix = "@";
        if((PRINT_LIST))
        {
            /* In list mode, when SECTIONS are not printed, provide   */
            /* these items as "pseudotags".                           */
            print_tag_address(ENTRY,start_offset,indent + SUBINDENT,prefix,addrwidth);
            printf("%-*.*s = %lu\n",TAGWIDTH + PREFIXLEN,
                    TAGWIDTH + PREFIXLEN,"Exif_MakerNoteCount",
                    maker_count ? maker_count: maker_entry->count);
            print_tag_address(ENTRY,start_offset,indent + SUBINDENT,prefix,addrwidth);
            printf("%-*.*s = ",TAGWIDTH + PREFIXLEN,
                    TAGWIDTH + PREFIXLEN,"Exif_MakerNoteScheme");
        }
        else
        {
            print_tag_address(SECTION,start_offset,indent + SUBINDENT,prefix,addrwidth);
            printf("<MakerNote> count %lu, ",
                    maker_count ? maker_count: maker_entry->count);
        }

        /* Discover, if possible, the 'scheme' by which the MakerNote */
        /* is stored.                                                 */
        maker_scheme = find_makerscheme(inptr,maker_offset + fileoffset_base,
                                            byteorder,maker_count,make,model);

        switch(maker_scheme->scheme_type)
        {
            case HAS_ID_SCHEME:
                if(make == MAKER_ASAHI)
                    printf("Asahi ");
                printf("ID");
                break;
            case HAS_ID_PLUS_TIFFHDR_SCHEME:
                printf("ID plus TIFFheader");
                break;
            case PLAIN_IFD_SCHEME:
                /* hack, hack, cough...'scuse me...                   */
                /* This really should be discovered in                */
                /* find_makerscheme()                                 */
                if(make == MAKER_ASAHI)
                {
                    printf("Note-relative ");
                    ++note_relative_offset_scheme;
                }
                else
                    printf("\"Normal ");
                printf("IFD");
                if(maker_entry->count == 4UL) /* some old Casio       */
                    printf(" with LONG offset\"");
                else
                    putchar('"');
                break;
            case FUJI_SCHEME:
                printf("FUJI");
                break;
            /* New schemes can be added here.  ###%%%                 */
            case UNKNOWN_SCHEME:
            default:
                printf("unknown");
                if(PRINT_SECTION)
                    printf(" scheme");
                else
                    putchar('\n');
                    goto dumpit;
                break;
        }
        if((PRINT_LIST))
            putchar('\n');
        else
            printf(" scheme,");
        num_entries = maker_scheme->num_entries;
        byteorder = maker_scheme->byteorder;
        if(maker_scheme->fileoffset_base)
        {
            fileoffset_base = maker_scheme->fileoffset_base;
            start_entry_offset = fileoffset_base + maker_scheme->ifd_offset + 2;
        }
        else
            start_entry_offset = start_offset + maker_scheme->ifd_offset + 2;
        if(PRINT_SECTION)
            printf(" %d entries starting at offset %#lx/%lu\n",num_entries,
                        start_entry_offset,start_entry_offset);

#ifdef SHOW_MAKERID
        extraindent(SUBINDENT);
        if(PRINT_SECTION)
            printf("{ Make_name=%s (ep maker #%d); Model_name=%s, Software_name=%s (ep model #%d) }\n",
                        Make_name?Make_name:"",make,Model_name?Model_name:"",
                        Software_name?Software_name:"",model);
#endif

        if((maker_scheme->id_length > 0) && (PRINT_SECTION))
        {
            print_tag_address(SECTION,start_offset, indent+SUBINDENT,
                                    prefix,addrwidth);
            extraindent(SUBINDENT);
            printf("MakerId=");
            (void)print_ascii(inptr,maker_scheme->id_length,start_offset);
            (void)putchar('\n');
        }

        if(((maker_scheme->fileoffset_base) &&
                (maker_scheme->scheme_type != FUJI_SCHEME)) &&
                    (PRINT_SECTION))
        {
            print_tag_address(SECTION,fileoffset_base,indent+SUBINDENT,
                                    prefix,addrwidth);
            extraindent(SUBINDENT);
            (void)print_header(maker_scheme->byteorder,SECTION);
            printf(" ifd offset = %#lx/%ld)\n",maker_scheme->ifd_offset,
                        maker_scheme->ifd_offset);
        }

        /* At this point, the note has been "verified" to be in TIFF  */
        /* IFD format, and can be read normally in that fashion.      */
        /* Maker-specific information is handled in the print_maker   */
        /* routines, according to the make and model numbers.         */

        /* Scheme validation is not entirely robust, however, so the  */
        /* code below bails out at the first sign of an invalid       */
        /* entry, and hexdumps the data if wanted.                    */

        current_offset = entry_offset = start_entry_offset;
        value_is_offset = 0;
        for(entry_num = 0; entry_num < num_entries; ++entry_num)
        {
            print_tag_address(ENTRY,entry_offset,indent+SUBINDENT,prefix,addrwidth);
            entry_ptr = read_entry(inptr,byteorder,entry_offset);
            if((entry_ptr->value_type == 0) || (entry_ptr->value_type > DOUBLE) ||
                ferror(inptr) || feof(inptr))
            if(invalid)
            {
                extraindent((addrwidth * 2) + 6 + indent + (SUBINDENT * 2));
                if(PRINT_SECTION)
                    printf("INVALID MAKERNOTE ENTRY %d at offset %lu\n",
                            invalid,entry_offset);
                goto dumpit;
            }
            if(ferror(inptr))
                goto dumpit;
            if(note_relative_offset_scheme)
                fileoffset_base = entry_offset;
            entry_offset = current_offset = ftell(inptr);
            value_offset = print_makerentry(inptr,byteorder,addrwidth,
                                            entry_ptr,fileoffset_base,
                                                indent,make,model);
            if(value_offset == 0UL)
                value_offset = entry_offset;
            if(value_offset > max_value_offset)
                max_value_offset = value_offset;
            if(is_offset(entry_ptr) && (PRINT_VALUE_AT_OFFSET))
                ++use_second_pass;
        }

        next_ifd_offset = read_ulong(inptr,byteorder,current_offset);
        if(next_ifd_offset > 0L)
        {
            /* This should not happen, of course.                     */
            if((PRINT_ADDRESS) && (PRINT_VALUE) && (PRINT_VALUE_AT_OFFSET))
            {
                print_tag_address(ENTRY,current_offset,indent+SUBINDENT,prefix,addrwidth);
                extraindent(SUBINDENT);
                printf("**** next IFD offset %lu  ",next_ifd_offset);
                next_ifd_offset += fileoffset_base;
                printf("(+ %lu = %#lx/%lu)\n",fileoffset_base,
                            next_ifd_offset,next_ifd_offset);
                if(next_ifd_offset < ftell(inptr))
                {
                    printred("BAD NEXT IFD OFFSET\n");
                    next_ifd_offset = 0L;
                }
                else
                    putchar('\n');
            }
            else
                next_ifd_offset += fileoffset_base;
        }
        else
        {
            if((PRINT_ADDRESS) && (PRINT_VALUE) && (PRINT_VALUE_AT_OFFSET))
            {
                print_tag_address(ENTRY,current_offset,indent+SUBINDENT,prefix,addrwidth);
                extraindent(SUBINDENT);
                printf("**** next IFD offset 0\n");
            }
        }

        if(ferror(inptr))
        {
            printf(" READ NEXT IFD OFFSET FAILED ");
            why(stdout);
            clearerr(inptr);    /* keep going...                      */
        }
        else
            current_offset = ftell(inptr);

        value_offset = current_offset;

        /* Make a second pass over the entries, to print values of    */
        /* entries whose values are indirect.                         */
        if(use_second_pass)
        {

            if((PRINT_VALUE) && (PRINT_ADDRESS) && (PRINT_VALUE_AT_OFFSET))
            {
                print_tag_address(ENTRY,value_offset,indent+SUBINDENT,prefix,addrwidth);
                extraindent(SUBINDENT);
                printf("============= VALUES, MakerNote ============\n");
            }

            entry_offset = start_entry_offset;
            for(entry_num = 0; entry_num < num_entries; ++entry_num)
            {
                entry_ptr = read_entry(inptr,byteorder,entry_offset);
                if(ferror(inptr))   /* why didn't this happen above?  */
                    goto dumpit;
                if(note_relative_offset_scheme)
                    fileoffset_base = entry_offset;
                entry_offset = current_offset = ftell(inptr);
                if(is_offset(entry_ptr))
                {
                    if((PRINT_VALUE_AT_OFFSET))
                        value_offset =
                            print_offset_makervalue(inptr,byteorder,
                                            entry_ptr,fileoffset_base,
                                            prefix,indent+SUBINDENT,
                                            addrwidth,make,model,1);
                    else
                        value_offset = fileoffset_base + entry_ptr->value + 
                            (value_type_size(entry_ptr->value_type) * entry_ptr->count);
                    if(value_offset > max_value_offset) 
                        max_value_offset = value_offset;
                }
            }
        }
        else if(current_offset > max_value_offset)
            max_value_offset = current_offset;

        /* If the note is incompletely specified, point that out. If  */
        /* IFD entries are not being printed this message is not      */
        /* wanted (and max_value_offset is likely 0 anyway).          */
        /* This actually happens...                                   */
        if((end_of_note && (max_value_offset < end_of_note)) && (PRINT_ENTRY) &&
            !(PRINT_LIST))
        {
            unsigned long printlength = 0;

            print_tag_address(ENTRY,max_value_offset - 1,indent + SUBINDENT,"*",addrwidth);
            extraindent(SUBINDENT);
            printf("---- End of values before end of Makernote\n");
            dumplength = end_of_note - max_value_offset;
            /* dump at least the first couple of rows, just to see... */
            if(PRINT_MAKERDUMP)
            {
                if((Max_makerdump == DUMPALL) || (Max_makerdump > dumplength))
                    printlength = dumplength;
                else if(Max_makerdump > 0L)
                    printlength = Max_makerdump;
                else if(Max_undefined > 0L)
                    printlength = Max_undefined;
            }
            else    /* always print a little bit in this case         */
                printlength = dumplength > 48 ? 48 : dumplength;
            hexdump(inptr,max_value_offset,dumplength,printlength,12,
                            indent,SUBINDENT,addrwidth);
            (void)putchar('\n');
        }
        else if(maker_count == 0L)
            end_of_note = max_value_offset;
            
        print_tag_address(SECTION,end_of_note - 1,indent + SUBINDENT,"-",addrwidth);
        if(max_offset && (start_offset > max_offset))
            end_of_note = 0L;
        if(PRINT_SECTION)
            printf("</MakerNote>");
        if(next_ifd_offset)
            printf("# WARNING! MakerNote has next ifd offset %lu\n",next_ifd_offset);
        else if(PRINT_SECTION)
            (void)putchar('\n');
        STOPCOLOR;
    }
    return(end_of_note);

dumpit:
    if(PRINT_MAKERDUMP)
    {
        /* If no specific length is given for makernote data, limit   */
        /* the dump length to the length specified for undefined      */
        /* data.                                                      */
        if((Max_makerdump == DUMPALL) || (Max_makerdump > maker_count))
            dumplength = maker_count;
        else if(Max_makerdump > 0L)
            dumplength = Max_makerdump;
        else
            dumplength = Max_undefined;
        if(dumplength > 0L)
        {
            (void)putchar('\n');
#ifdef SHOW_MAKERID
            extraindent(SUBINDENT);
            printf("{ Make_name=%s (ep maker #%d); Model_name=%s, Software_name=%s (ep model #%d) }\n",
                        Make_name?Make_name:"",make,Model_name?Model_name:"",
                        Software_name?Software_name:"",model);
#endif
            hexdump(inptr,start_offset,maker_count,dumplength,12,
                    indent,SUBINDENT,addrwidth);
            (void)putchar('\n');
        }
    }
    else
    {
        if(PRINT_SECTION)
            printred(" (not dumped: use -M)\n");
#ifdef SHOW_MAKERID
            extraindent(SUBINDENT);
        if(PRINT_SECTION)
            printf("{ Make_name=%s (ep maker #%d); Model_name=%s, Software_name=%s (ep model #%d) }\n",
                        Make_name?Make_name:"",make,Model_name?Model_name:"",
                        Software_name?Software_name:"",model);
#endif
    }
    if(PRINT_SECTION)
    {
        print_tag_address(SECTION,end_of_note - 1,indent + SUBINDENT,"-",addrwidth);
        printf("</MakerNote>\n");
    }
    STOPCOLOR;
    if(maker_count)
        return(end_of_note);
    return(0L);
cant_dumpit:
    printf(" INVALID MakerNote offset %lu\n",maker_offset);
    return(0L);
}

/* Return a unique number for a given maker, (assigned in "maker.h")  */
/* based upon a string, usually taken from a "Make" tag. This "magic  */
/* number" will be used to select routines to print and interpret     */
/* data from MakerNotes for that maker.                               */

int
maker_number(char *make)
{
    int number = GENERIC;

    if(make)
    {
        /* Fast first match, then string compare                      */
        switch(*make)
        {
            case 'a':
            case 'A':
                if(strncasecmp(make,"agfa",4) == 0)
                    number = MAKER_AGFA;
                else if(strncasecmp(make,"asahi",5) == 0)
                    number = MAKER_ASAHI;   /* Pentax                 */
                /* else add makers here                               */
                break;
            case 'c':
            case 'C':
                if(strncasecmp(make,"canon",5) == 0)
                    number = MAKER_CANON;
                else if(strcasecmp(make,"casio") == 0)
                    number = MAKER_CASIO;
                /* else add makers here                               */
                break;
            case 'e':
            case 'E':
                if(strncasecmp(make,"eastman",7) == 0)
                    number = MAKER_KODAK;
                /* else add makers here                               */
                break;
            case 'f':
            case 'F':
                if(strncasecmp(make,"fuji",4) == 0)
                    number = MAKER_FUJIFILM;
                /* else add makers here                               */
                break;
            case 'h':
            case 'H':
                if(strncasecmp(make,"hewlett",7) == 0)
                    number = MAKER_HP;
                /* else add makers here                               */
                break;
            case 'k':
            case 'K':
                if(strncasecmp(make,"konica",6) == 0)
                    number = MAKER_KONICA;
                else if(strncasecmp(make,"kyocera",7) == 0)
                    number = MAKER_KYOCERA;
                /* else add makers here                               */
                break;
            case 'l':
            case 'L':
                if(strncasecmp(make,"leica",5) == 0)
                    number = MAKER_LEICA;
                /* else add makers here                               */
                break;
            case 'm':
            case 'M':
                if(strncasecmp(make,"minolta",7) == 0)
                    number = MAKER_MINOLTA;
                /* else add makers here                               */
                break;
            case 'p':
            case 'P':
                if(strncasecmp(make,"panasonic",9) == 0)
                    number = MAKER_PANASONIC;
                /* else add makers here                               */
                break;
            case 'n':
            case 'N':
                if(strncasecmp(make,"nikon",5) == 0)
                    number = MAKER_NIKON;
                /* else add makers here                               */
                break;
            case 'o':
            case 'O':
                if(strncasecmp(make,"olympus",7) == 0)
                    number = MAKER_OLYMPUS;
                /* else add makers here                               */
                break;
            case 'r':
            case 'R':
                if(strncasecmp(make,"ricoh",5) == 0)
                    number = MAKER_RICOH;
                /* else add makers here                               */
                break;
            case 's':
            case 'S':
                if(strncasecmp(make,"sony",4) == 0)
                    number = MAKER_SONY;
                else if(strncasecmp(make,"sanyo",5) == 0)
                    number = MAKER_SANYO;
                else if(strncasecmp(make,"seiko",5) == 0)
                    number = MAKER_EPSON;
                /* else add makers here                               */
                break;
            case 't':
            case 'T':
                if(strncasecmp(make,"toshiba",8) == 0)
                    number = MAKER_TOSHIBA;
                if(strncasecmp(make,"traveler",8) == 0)
                    number = MAKER_TRAVELER;
                /* else add makers here                                   */
                break;
            default:
                break;
        }
    }
    return(number);
}

/* Call a maker-specific routine to assign an identifying number for  */
/* a specific model of a given maker's camera. The camera make is     */
/* identified by the number assigned by make_number(), and is used to */
/* dispatch the maker-specific routine, which is then free to pick an */
/* identifying number based upon the "Model" tag and possibly the     */
/* "Software" tag. The numbers are assigned in "maker.h" and will be  */
/* used to select print routines appropriate to the makernote data    */
/* for the current image. It is possible to assign a single number to */
/* a group of models, or to assign a separate number to each          */
/* model/software combination, requiring the maker-specific routines  */
/* to dispatch the appropriate routines.                              */

int
model_number(int make,char *model,char *software)
{
    int number = GENERIC;

    if(model == NULL)
        model = "";
    if(software == NULL)
        software = "";
    switch(make)
    {
        case MAKER_AGFA:
            number = agfa_model_number(model,software);
            break;
        case MAKER_ASAHI:   /* PENTAX                                 */
            number = asahi_model_number(model,software);
            break;
        case MAKER_CANON:
            number = canon_model_number(model,software);
            break;
        case MAKER_CASIO:
            number = casio_model_number(model,software);
            break;
        case MAKER_EPSON:
            number = epson_model_number(model,software);
            break;
        case MAKER_FUJIFILM:
            number = fujifilm_model_number(model,software);
            break;
        case MAKER_HP:      /* no makernote data                      */
            break;
        case MAKER_KODAK:   /* no makernote data                      */
            break;
        case MAKER_KONICA:  /* no makernote data                      */
            break;
        case MAKER_KYOCERA: /* no makernote data                      */
            break;
        case MAKER_LEICA:
            number = leica_model_number(model,software);
            break;
        case MAKER_MINOLTA:
            number = minolta_model_number(model,software);
            break;
        case MAKER_NIKON:
            number = nikon_model_number(model,software);
            break;
        case MAKER_OLYMPUS:
            number = olympus_model_number(model,software);
            break;
        case MAKER_PANASONIC: /* no makernote data                    */
            break;
        case MAKER_RICOH:    /* no makernote data                     */
            break;
        case MAKER_SANYO:
            number = sanyo_model_number(model,software);
            break;
        case MAKER_SONY:    /* no makernote data                      */
            break;
        case MAKER_TOSHIBA:     /* no model info                      */
            break;
        case MAKER_TRAVELER:    /* no model info                      */
            break;
        default:    /* GENERIC                                        */
            break;
    }
    return(number);
}

/* Check the beginning of a MakerNote to determine if it begins with  */
/* an ascii string, presumed to identify the maker (again).           */

int
check_makernote(FILE* inptr,unsigned long offset)
{
    char *name;
    int length = 0;

    if(inptr && offset)
    {
        name = (char *)read_bytes(inptr,12,offset);
        if(name)
        {
            length = strlen(name);
            if(length <= 4)
            {
                /* Heurism is all well and good, but this is based    */
                /* upon just one found image from a Pentax            */
                /* Optio430RS, and could be a red herring.            */
                if((length == 3) && (strncmp(name,"AOC",3) == 0))
                    length = 6;
                else
                    length = 0;
            }
            else if(length >= 8)
            {
                if(strncmp(name,"FUJI",4) == 0)
                    length = -12;   /* special case                   */
                else
                    length = 12;
            }
            else
                length = 8;
        }
        else
            length = -1;
    }
    return(length);
}

/* Attempt to discover the format of a MakerNote. Some notes are in   */
/* normal TIFF IFD format, with tags assigned by the maker, and       */
/* values interpreted in maker-specific ways. Some makers begin with  */
/* a few bytes containing an ascii string identifying the maker, with */
/* a normal IFD following at a specific offset. At least one camera   */
/* I've seen begins with such an ascii id, followed by an actual TIFF */
/* header, with the following IFD offsets then relative to that TIFF  */
/* header. Finally, one maker uses an ascii id string, as above, but  */
/* writes offsets in the following IFD relative to the start of the   */
/* note rather than the TIFF header, and always uses TIFF_INTEL byte  */
/* order, regardless the byte order of the remainder of the file.     */

/* This routine uses some simple heuristics to attempt to intuit the  */
/* type of note, rather than a table of known make/model formats, so  */
/* that it has some chance of presenting the structure of notes the   */
/* program has never seen before. If the note just doesn't look       */
/* presentable, a (currently dummy) routine will be used to check for */
/* known sports.                                                      */

/* Note that even "known" makernotes must pass this test first (and   */
/* will if not corrupted). The print routines will then have a chance */
/* to intercept and interpret any of the freakish tags that some      */
/* makers like to use.                                                */


struct maker_scheme *
find_makerscheme(FILE *inptr,unsigned long start_offset,unsigned short byteorder,
                    unsigned long maker_size,int make,int model)
{
    static struct maker_scheme maker_scheme;
    int id_length,num_entries;
    unsigned long ifd_offset = 0L;
    unsigned long fileoffset_base = 0L;
    unsigned long scheme_type = UNKNOWN_SCHEME;
    unsigned short marker;

    id_length = check_makernote(inptr,start_offset);
#ifdef DEBUG_SCHEME
    pushcolor(GREEN);
    printf("\nScheme:\nInitial id length = %d",id_length);
    popcolor();
#endif
    switch(id_length)
    {
        case -1:    /* read failed                                    */
        case 0:     /* normal or unknown                              */
            break;
        case 6:
            ifd_offset = 6; /* Asahi strikes again...                 */
            scheme_type = HAS_ID_SCHEME;
            break;
        case 8:
            ifd_offset = 8;
            scheme_type = HAS_ID_SCHEME;
            marker = read_imagemarker(inptr,start_offset + 10);
            switch(marker)
            {
                case TIFF_INTEL:
                case TIFF_MOTOROLA:
                    scheme_type = HAS_ID_PLUS_TIFFHDR_SCHEME;
                    byteorder = marker;
                    id_length = 10;   /* only Nikon D100 so far     */
                    fileoffset_base = start_offset + id_length;
                    start_offset = 0L;
                    ifd_offset = read_ulong(inptr,byteorder,HERE);
                    break;
            }
            break;
        case 12:
            ifd_offset = 12;
            scheme_type = HAS_ID_SCHEME;
            break;
        case -12:
            byteorder = TIFF_INTEL;
            id_length = ifd_offset = 12;
            fileoffset_base = start_offset;
            start_offset = 0L;
            scheme_type = FUJI_SCHEME;
            break;
        default:
            break;
    }
#ifdef DEBUG_SCHEME
    pushcolor(GREEN);
    printf(" -> %d\n",id_length);
    printf("IFD offset = %lu\n",ifd_offset);
    printf("Scheme type = %#lx\n",scheme_type);
    popcolor();
#endif

    if(id_length >= 0)
        num_entries = read_ushort(inptr,byteorder,start_offset + ifd_offset + fileoffset_base);
    else
        num_entries = 0;
#ifdef DEBUG_SCHEME
    pushcolor(GREEN);
    printf("num entries = %d\n",num_entries);
    popcolor();
#endif

    /* This catches most, but not necessarily all...                  */
    if(((maker_size > 0L) && (num_entries > (maker_size / 12))) || (num_entries <= 1))
        find_maker_model(&maker_scheme,make,model);
    else
    {
        /* We seem to have a winner...                                */
        if(scheme_type == UNKNOWN_SCHEME)
            maker_scheme.scheme_type = PLAIN_IFD_SCHEME;
        else
            maker_scheme.scheme_type = scheme_type;
        maker_scheme.fileoffset_base = fileoffset_base;
        maker_scheme.ifd_offset = ifd_offset;
        maker_scheme.id_length = id_length;
        maker_scheme.num_entries = num_entries;
        maker_scheme.byteorder = byteorder;
#ifdef DEBUG_SCHEME
    pushcolor(GREEN);
    printf("Scheme type => %#lx\n",maker_scheme.scheme_type);
    printf("IFD offset = %lu\n",maker_scheme.ifd_offset);
    printf("id length = %ld\n",maker_scheme.id_length);
    printf("fileoffset base = %ld\n",maker_scheme.fileoffset_base);
    printf("num entries = %d\n",maker_scheme.num_entries);
    printf("byteorder = %#x\n",maker_scheme.byteorder);
    popcolor();
#endif

    }
    return(&maker_scheme);
}

/* Look up a make and model in a table of known sports. This is       */
/* currently just a dummy, but might be used to handle non-ifd notes  */
/* by setting a new type here and revising the generic code to        */
/* dispatch a specialized routine. I have no information on such      */
/* notes now.                                                         */

void
find_maker_model(struct maker_scheme *maker_scheme,int make,int model)
{
        /* Initialize the structure                                   */
        maker_scheme->scheme_type = UNKNOWN_SCHEME;
        maker_scheme->ifd_offset = 0L;
        maker_scheme->fileoffset_base = 0L;
        maker_scheme->id_length = 0;
        maker_scheme->num_entries = 0;
        maker_scheme->byteorder = 0;
        /* do a lookup here on make and model                         */
}
