/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: maker_minolta.c,v 1.18 2003/08/05 16:58:29 alex Exp $";
#endif

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* minolta camera maker-specific routines                             */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Most of the information used here is due to Dalibor Jelinek at:    */
/*     http://www.dalibor.cz/minolta/makernote.htm                    */
/*                                                                    */
/* The F100 and "X" models *appear* to use the same MakerNote tagset  */
/* as used for Sanyo models, as decoded by John Hawkins               */
/*     http://www.exif.org/makernotes/SanyoMakerNote.html             */
/*                                                                    */
/* NOTE that this is guesswork. Test images I have found for these    */
/* models appear to decode properly using the tag numbers from        */
/* Hawkinn's Sanyo data.  Send mail if you can confirm or deny.       */
/*                                                                    */
/* The EX model MakerNote format is indecipherable.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "maker.h"

extern void minolta_camera_setting(FILE *,unsigned short,int,
                                struct ifd_entry *,unsigned long,int);

/* Find the identifying number assigned to known Minolta camera       */
/* models. This number is used to dispatch print and interpret        */
/* routines approopriate to the current image.                        */

int
minolta_model_number(char *model,char *software)
{
    int number = GENERIC;

    if(model)
    {
        switch(*model)
        {
            case 'D':
                if(strcmp(model,"DiMAGE 7Hi") == 0)
                    number = MINOLTA_DIMAGE7Hi;
                else if(strcmp(model,"DiMAGE 7i") == 0)
                    number = MINOLTA_DIMAGE7i;
                else if(strcmp(model,"DiMAGE 7") == 0)
                    number = MINOLTA_DIMAGE7;
                else if(strcmp(model,"DiMAGE F100") == 0)
                    number = MINOLTA_DIMAGEF100;
                else if(strcmp(model,"DiMAGE F300") == 0)
                    number = MINOLTA_DIMAGEF300;
                else if(strcmp(model,"DiMAGE S404") == 0)
                    number = MINOLTA_DIMAGES404;
                break;
            default:
                number = GENERIC;
                break;
        }
    }
    return(number);
}


/* Dispatch a print routine based upon model                          */

void
print_minolta_makervalue(struct ifd_entry *entry_ptr,int make, int model,
                    char *prefix)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            case MINOLTA_DIMAGE7:
            case MINOLTA_DIMAGE7i:
            case MINOLTA_DIMAGE7Hi:
            case MINOLTA_DIMAGEF100:    /* sanyo                      */
            case MINOLTA_DIMAGEF300:
            case MINOLTA_DIMAGES404:
            case MINOLTA_DIMAGEEX:      /* indecipherable             */
            case MINOLTA_DIMAGEX:       /* sanyo                      */
            case GENERIC:
            default:
                print_minolta1_makervalue(entry_ptr,make,model,prefix);
                minolta1_interpret_value(entry_ptr,model);
                break;
        }
    }
}

/* Model-specific print routine for minolta cameras. This routine is  */
/* responsible for picking off any direct entry tags which are        */
/* peculiar and will not be handled properly by print_value()         */
/* (usually UNDEFINED values which fit in the 4-byte entry value). If */
/* there are no such entries, this function simply calls              */
/* print_value().                                                     */

void
print_minolta1_makervalue(struct ifd_entry *entry_ptr,int make, int model,
                    char *prefix)
{
    if(entry_ptr && (PRINT_VALUE))
    {
        switch(entry_ptr->tag) 
        {
                default:
                print_value(entry_ptr,PREFIX);
                break;
        }
    }
}

/* Dispatch a routine to decode and print offset values for minolta   */
/* cameras.                                                           */
void
print_minolta_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            default:
                minolta1_offset_makervalue(inptr,byteorder,entry_ptr,
                                            offset_ref,prefix,indent,
                                            addrwidth,make,model,at_offset);
                minolta1_interpret_offset_makervalue(inptr,byteorder,entry_ptr,
                                            offset_ref);
                break;
        }
    }
}

/* Model-specific routine to print UNDEFINED values found at offsets  */
/* in minolta makernotes. One of these may be supplied for each model */
/* with unique UNDEFINED tags.                                        */

void
minolta1_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{
    unsigned long value_offset;
    unsigned long dumplength;
    unsigned short marker;
    char *nameoftag;

    if(entry_ptr)
    {
        nameoftag = maker_tagname(entry_ptr->tag,make,model); 
        value_offset = offset_ref + entry_ptr->value;
        switch(entry_ptr->tag)
        {
            case 0x0001:    /* Camera Settings, old                   */
            case 0x0003:    /* Camera Settings, new                   */
                if(at_offset)
                {
                    print_tag_address(ENTRY,value_offset,indent,prefix,addrwidth);
                    print_makertagid(entry_ptr,14,": ",SUBINDENT,make,model);
                }
                printf("%lu entries",entry_ptr->count / 4);
                minolta_camera_setting(inptr,byteorder,model,
                                entry_ptr,value_offset,indent + SUBINDENT);
                if(PRINT_SECTION)
                {
                    print_tag_address(SECTION,value_offset + (entry_ptr->count * 2) - 1,
                                    indent,prefix,addrwidth);
                    extraindent(SUBINDENT);
                    printf("End of %s",nameoftag);
                }
                break;
            case 0x0081:    /* JPEG Thumbnail in TIFF MakerNote       */
                if(at_offset)
                {
                    print_tag_address(VALUE_AT_OFFSET,value_offset,indent,prefix,addrwidth);
                    extraindent(SUBINDENT);
                    printf("Start of JPEG Thumbnail");
                }
                if(!(PRINT_OFFSET))
                {
                    if(PRINT_BOTH_VALUE)
                        printf("@%#lx=%lu",value_offset,value_offset);
                    else if(PRINT_HEX_VALUE)
                        printf("@%#lx",value_offset);
                    else if(PRINT_DEC_VALUE)
                        printf("@%lu",value_offset);
                }
                printf(" length %ld",entry_ptr->count);
                marker = read_imagemarker(inptr,value_offset);
                if(PRINT_SEGMENT)
                {
                    putchar('\n');
                    (void)process_jpeg_segments(inptr,value_offset,
                                                marker,entry_ptr->count,
                                                indent+(SUBINDENT * 2),addrwidth);
                    print_tag_address(SEGMENT,value_offset + entry_ptr->count - 1,
                                                indent,"-",addrwidth);
                    extraindent(SUBINDENT);
                    printf("End of JPEG Thumbnail");
                }
                /* make certain we're at the end unless 'inptr' bad   */
                fseek(inptr,value_offset + entry_ptr->count,0);
                break;
            case 0x0e00: /* PrintIM                                   */
            case 0x0f00: /* Data                                      */
                if(at_offset)
                {
                    print_tag_address(ENTRY,value_offset,indent,prefix,addrwidth);
                    print_makertagid(entry_ptr,19," : ",SUBINDENT,make,model);
                }
                else if(PRINT_OFFSET)
                        printf(" = ");
                printf("UNDEFINED, count %lu", entry_ptr->count);
                if(Max_undefined == 0)
                    printred(" (not dumped, use -U)");
                else
                {
                    if((Max_undefined == DUMPALL)
                        || (Max_undefined > entry_ptr->count))
                            dumplength = entry_ptr->count;
                    else 
                        dumplength = Max_undefined;
                    putchar('\n');
                    hexdump(inptr,value_offset,entry_ptr->count,dumplength,12,
                                indent,SUBINDENT,addrwidth);
                }
                /* make certain we're at the end                      */
                fseek(inptr,value_offset + entry_ptr->count,0);
                break;
            default:
                print_generic_offset_makervalue(inptr,byteorder,entry_ptr,
                                        offset_ref,prefix,indent,
                                        addrwidth,make,model,at_offset);
                break;
        }
    }
}

/* minolta-specific tagnames for makernotes.                          */

char *
maker_minolta_tagname(unsigned short tag,int model)
{
    char *tagname = (char *)0;

    /* Check specific models first                                    */
    switch(model)
    {
        /* Add special-case models here                               */
        default:
            break;
    }

    /* The Minolta MakerNote of TIFF format images may have a JPEG    */
    /* thumbnail with a defective SOI marker (0x00f8) based upon ONE  */
    /* image I've seen.                                               */

    if(tagname == NULL)
    {
        switch(tag)
        {
            case 0x0000: tagname = "Version"; break;
            case 0x0001: tagname = "CameraSettings"; break;
            case 0x0003: tagname = "CameraSettings"; break;
            case 0x0040: tagname = "ImageByteSize"; break;
            case 0x0081: tagname = "JPEG_Thumbnail"; break;
            case 0x0088: tagname = "JpegThumbnailOffset"; break;    /* ###%%% print it */
            case 0x0089: tagname = "JpegThumbnailLength"; break;    /* ###%%% print it */
            case 0x0101: tagname = "ColorMode"; break;
            case 0x0102: tagname = "ImageQuality"; break;
            case 0x0103: if(model == MINOLTA_DIMAGE7Hi)
                              tagname = "ImageQuality";
                         else
                              tagname = "ImageSize";
                         break;
            case 0x0200: tagname = "SpecialMode"; break;

    /* The following tag names appear valid for the F100 and "X" but  */
    /* do not appear to be used for any other Minolta model. They are */
    /* essentially the Sanyo tag set, used here based upon guesswork; */
    /* some Minolta and Sanyo cameras are said to use the same chip,  */
    /* and some F100 and "X" images I've seen contain almost exactly  */
    /* this set of tags (as described by John Hawkins for Sanyo       */
    /* DSC-MZ2 notes). The Quality, Macro, SpecialMode and PrintIM    */
    /* tags are reasonably certain to be correct (they "decode"       */
    /* properly). The remaining names are less certain, and no        */
    /* attempt is made to decode them in minolta1_interpret_value()/  */
            case 0x0201: tagname = "Quality"; break;
            case 0x0202: tagname = "Macro"; break;
            case 0x0204: tagname = "DigiZoom"; break;
            case 0x0207: tagname = "SoftwareRelease"; break;
            case 0x0208: tagname = "PictInfo"; break;
            case 0x0209: tagname = "Camera ID"; break;
            case 0x020e: tagname = "SeqShotMethod"; break;
            case 0x020f: tagname = "WideRange"; break;
            case 0x0210: tagname = "ColorAdjMode"; break;
            case 0x0213: tagname = "QuickShot"; break;
            case 0x0214: tagname = "SelfTimer"; break;
            case 0x0216: tagname = "VoiceMemo"; break;
            case 0x0217: tagname = "RecordShutterRel"; break;
            case 0x0218: tagname = "FlickerReduce"; break;
            case 0x0219: tagname = "OpticalZoom"; break;
            case 0x021b: tagname = "DigitalZoom"; break;
            case 0x021d: tagname = "LightSourceSpecial"; break;
            case 0x021e: tagname = "Resaved"; break;
            case 0x021f: tagname = "SceneSelect"; break;
            case 0x0223: tagname = "ManualFocalDist"; break; /* ###%%% not seen */
            case 0x0224: tagname = "SeqShotIntvl"; break; /* ###%%% not seen */
            case 0x0225: tagname = "FlashMode"; break; /* ###%%% not seen */
            case 0x0e00: tagname = "PrintIMData"; break;
            case 0x0f00: tagname = "Data"; break;
            default: break;
        }
    }
    return(tagname);
}


void
minolta1_interpret_value(struct ifd_entry *entry_ptr,int model)
{
    switch(entry_ptr->tag)
    {
        case 0x0101:    /* ColorMode                                  */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Natural"); break;
                case 1: printf("Black/White"); break;
                case 2: printf("Vivid"); break;
                case 3: printf("Solarize"); break;
                case 4: printf("AdobeRGB"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0102:    /* CameraSettings ImageQuality, 7, 7Hi, F300  */
minolta_7Hi_botch:
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Raw"); break;
                case 1: printf("Uncompressed"); break;
                case 2: printf("Fine(1/2)"); break;
                case 3: printf("Normal(1/3)"); break;   /* ###%%% */
                case 4: printf("Economy(1/4)"); break;  /* ###%%% */
                case 5: printf("ExtraFine(1/1)"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0103:    /* CameraSettings ImageSize*, 7Hi, F300       */
            /* The 7Hi appears to get this wrong, printing the same   */
            /* value for this as for 0x0102; the F300 seems to do it  */
            /* right. The ImageSize entry in "Camera Settings"        */
            /* appears correct in both cases. For the 7Hi, it would   */
            /* be wrong to report this as image size, so just don't   */
            /* do it.                                                 */
            if(model == MINOLTA_DIMAGE7Hi)
                goto minolta_7Hi_botch;

            print_startvalue();
            switch(entry_ptr->value)
            {
                case 1: printf("1600x1200"); break;
                case 2: printf("unknown"); break;   /* 1280x960? */
                case 3: printf("640x480"); break;
                case 4: printf("unknown"); break;
                case 5: printf("2560x1920"); break;
                case 6: printf("2272x1704"); break; 
                case 7: printf("2048x1536"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0201:    /* Quality, X, F100                           */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Raw*"); break;
                case 1: printf("Economy(1/8)"); break;
                case 2: printf("Normal(1/4)"); break; 
                case 3: printf("Fine(1/2)"); break;
                case 4: printf("Uncompressed"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x0202:    /* Macro */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Normal"); break;
                case 1: printf("Macro"); break;
                case 2: printf("View"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        default:
            break;
    }
}

void
minolta1_interpret_offset_makervalue(FILE *inptr,unsigned short byteorder,
                    struct ifd_entry *entry_ptr,unsigned long offset_ref)
{
    unsigned long offset,value;

    if(entry_ptr)
    {
        offset = entry_ptr->value + offset_ref;
        switch(entry_ptr->tag)
        {
            case 0x0200:    /* SpecialMode                               */
                value = read_ulong(inptr,byteorder,offset);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Normal"); break;
                    case 1: printf("Unkown"); break;
                    case 2: printf("Fast"); break;
                    case 3: printf("Panorama,");
                            value = read_ulong(inptr,byteorder,HERE);
                            printf("#%lu,",value);
                            value = read_ulong(inptr,byteorder,HERE);
                            switch(value)
                            {
                                case 1: printf(" Left to Right"); break;
                                case 2: printf(" Right to Left"); break;
                                case 3: printf(" Bottom to Top"); break;
                                case 4: printf(" Top to Bottom"); break;
                                default: printred(" undefined"); break;
                            }
                            break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            default:
                break;
        }
    }
}

/* Interpret the "Camera Settings" tag data from Minolta MakerNotes.  */

#include <math.h>

void
minolta_camera_setting(FILE *inptr,unsigned short byteorder,int model,
    struct ifd_entry *entry_ptr,unsigned long offset,int indent)
{
    unsigned long value;
    double fvalue;
    int i,entries;

    if(PRINT_SECTION)
        printf(", length %lu",entry_ptr->count);
    putchar('\n');
    entries = entry_ptr->count / 4;
    for(i = 1; i <= entries; ++i)
    {
        value = read_ulong(inptr,TIFF_MOTOROLA,offset);
        if(PRINT_ADDRESS)
        {
            print_tag_address(ENTRY,offset,indent,"@",ADDRWIDTH);
            extraindent(SUBINDENT);
        }
        else
            printf("Exif_Maker_");
        offset += 4;
        switch(i)
        {
            case 1:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Unknown",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Unknown",value);
                break;
            case 2:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Exposure Mode",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"ExposureMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Programmed Auto"); break;
                    case 1: printf("Aperture Priority Auto"); break;
                    case 2: printf("Shuttter Priority Auto"); break;
                    case 3: printf("Manual"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 3:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Flash Mode",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"FlashMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("fill"); break;
                    case 1: printf("Redeye Reduction"); break;
                    case 2: printf("Rear Flash Sync"); break;
                    case 3: printf("Wireless"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 4:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"White Balance",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"WhiteBalance",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Auto"); break;
                    case 1: printf("Daylight"); break;
                    case 2: printf("Cloudy"); break;
                    case 3: printf("Tungsten"); break;
                    case 5: printf("Custom"); break;
                    case 7: printf("Fluoreschent"); break;
                    case 8: printf("Fluoreschent 2"); break;
                    case 11: printf("Custom 2"); break;
                    case 12: printf("Custom 3"); break;
                    default: printred("unknown"); break;
                }
                print_endvalue();
                break;
            case 5:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Image Size",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"ImageSize",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("2560x1920"); break;
                    case 1: printf("1600x1200"); break;
                    case 2: printf("1280x960"); break;
                    case 3: printf("640x480"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 6:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Image Quality",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"ImageQuality",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Raw"); break;
                    case 1: printf("SuperFine"); break;
                    case 2: printf("Fine"); break;
                    case 3: printf("Standard"); break;
                    case 4: printf("Economy"); break;
                    case 5: printf("ExtraFine"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 7:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Drive Mode",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"DriveMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Single"); break;
                    case 1: printf("Continuous"); break;
                    case 2: printf("Self Timer"); break;
                    case 4: printf("Bracketing"); break;
                    case 5: printf("Interval"); break;
                    case 6: printf("UHS Continuous"); break;
                    case 7: printf("HS Continuous"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 8:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Metering Mode",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"MeteringMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Multi-segment"); break;
                    case 1: printf("Center weighted"); break;
                    case 2: printf("Spot"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 9:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s",i,"Film Speed");
                else
                    printf("CS_%02d_%-21.21s",i,"FilmSpeed");
                if(PRINT_RAW_VALUES)
                    printf(" = %lu",value);
                fvalue = ((double)value/8.0) - 1.0;
                printf(" = %.3g APEX",fvalue);
                fvalue = pow(2.0,fvalue) * 3.125;
                print_startvalue();
                printf("%.3g ISO",fvalue);
                print_endvalue();
                break;
            case 10:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s",i,"Shutter Speed");
                else
                    printf("CS_%02d_%-21.21s",i,"ShutterSpeed");
                if(PRINT_RAW_VALUES)
                    printf(" = %lu",value);
                fvalue = ((double)value / 8.0) - 6.0;
                printf(" = %.3g APEX",fvalue);
                fvalue = pow(2.0,-fvalue);
                print_startvalue();
                printf("%.3g sec",fvalue);
                print_endvalue();
                break;
            case 11:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s",i,"Aperture");
                else
                    printf("CS_%02d_%-21.21s",i,"Aperture");
                if(PRINT_RAW_VALUES)
                    printf(" = %lu",value);
                fvalue = (double)value/8.0 - 1.0;
                printf(" = %.3g APEX",fvalue);
                fvalue = pow(2.0,fvalue/2.0);
                print_startvalue();
                printf("f%.2g",fvalue);
                print_endvalue();
                break;
            case 12:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Macro Mode",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"MacroMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("On"); break;
                    case 1: printf("Off"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 13:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Digital Zoom",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"DigitalZoom",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Off"); break;
                    case 1: printf("Electronic"); break;
                    case 0xffff: printf("2x"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 14:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Exposure Compensation",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"ExposureCompensation",value);
                print_startvalue();
                printf("%g EV",(double)value/3.0 - 2);
                print_endvalue();
                break;
            case 15:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Bracket Step",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"BracketStep",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("1/3"); break;
                    case 1: printf("2/3"); break;
                    case 2: printf("1"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 16:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Unknown",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Unknown",value);
                break;
            case 17:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Interval Length",value + 1);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"IntervalLength",value + 1);
                printf(" minutes");
                break;
            case 18:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Interval Number",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"IntervalNumber",value);
                break;
            case 19:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Focal Length",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"FocalLength",value);
                fvalue = (double)value/256.0;
                print_startvalue();
                printf("%g mm",fvalue);
                print_endvalue();
                break;
            case 20:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Focal Distance",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"FocalDistance",value);
                printf(" mm");
                break;
            case 21:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Flash",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Flash",value);
                print_startvalue();
                if(value == 0UL)
                    printf("did not fire");
                else
                    printf("fired");
                print_endvalue();
                break;
            case 22:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %#010lx",i,"Date",value);
                else
                    printf("CS_%02d_%-21.21s = %#010lx",i,"Date",value);
                print_startvalue();
                printf("%02ld:%02ld:%02ld",(value & 0xffff0000) >> 16,
                        (value & 0xff00) >> 8,value & 0xff);
                print_endvalue();
                break;
            case 23:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %#010lx",i,"Time",value);
                else
                    printf("CS_%02d_%-21.21s = %#010lx",i,"Time",value);
                print_startvalue();
                printf("%02ld:%02ld:%02ld",(value & 0xffff0000) >> 16,
                        (value & 0xff00) >> 8,value & 0xff);
                print_endvalue();
                break;
            case 24:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s",i,"Max Aperture");
                else
                    printf("CS_%02d_%-21.21s",i,"MaxAperture");
                if(PRINT_RAW_VALUES)
                    printf(" = %lu",value);
                fvalue = (double)value/8.0 - 1.0;
                printf(" = %.3g APEX",fvalue);
                fvalue = pow(2.0,fvalue/2.0);
                print_startvalue();
                printf("f%.2g",fvalue);
                print_endvalue();
                break;
            case 25:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Unknown",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Unknown",value);
                break;
            case 26:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Unknown",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Unknown",value);
                break;
            case 27:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"File Number Memory",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"FileNumberMemory",value);
                print_startvalue();
                if(value == 0UL)
                    printf("Off");
                else
                    printf("On");
                print_endvalue();
                break;
            case 28:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Last FileNumber",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"LastFileNumber",value);
                break;
            case 29:
                if(PRINT_SECTION)
                    printf("CS_%02d_%-21.21s = %lu",i,"White Balance Red",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"WhiteBalanceRed",value);
                print_startvalue();
                printf("%.3f",value/256.0);
                print_endvalue();
                break;
            case 30:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"White Balance Green",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"WhiteBalanceGreen",value);
                print_startvalue();
                printf("%.3f",value/256.0);
                print_endvalue();
                break;
            case 31:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"White Balance Blue",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"WhiteBalanceBlue",value);
                print_startvalue();
                printf("%.3f",value/256.0);
                print_endvalue();
                break;
            case 32:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Saturation",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Saturation",value);
                break;
            case 33:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Contrast",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Contrast",value);
                break;
            case 34:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Sharpness",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Sharpness",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Hard"); break;
                    case 1: printf("Normal"); break;
                    case 2: printf("Soft"); break;
                    case 3: printf("None"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 35:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Subject Program",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"SubjectProgram",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("None"); break;
                    case 1: printf("Portrait"); break;
                    case 2: printf("Text"); break;
                    case 3: printf("Night Portrait"); break;
                    case 4: printf("Sunset"); break;
                    case 5: printf("Sports Action"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 36:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Flash Compensation",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"FlashCompensation",value);
                fvalue = ((double)value - 6.0) / 3.0;
                print_startvalue();
                printf("%g EV",fvalue);
                print_endvalue();
                break;
            case 37:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"ISO Setting",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"ISOSetting",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("100"); break;
                    case 1: printf("200"); break;
                    case 2: printf("200"); break;
                    case 3: printf("800"); break;
                    case 4: printf("Auto"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 38:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Camera Model",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"CameraModel",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("DiMage 7"); break;
                    case 1: printf("DiMage 5"); break;
                    case 2: printf("DiMage S304"); break;
                    case 3: printf("DiMage S404"); break;
                    case 4: printf("DiMage 7i"); break;
                    case 5: printf("DiMage 7Hi"); break;
                    case 7: printf("DiMage S414"); break;
                    default: printred("unknown"); break;
                }
                print_endvalue();
                break;
            case 39:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Interval Mode",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"IntervalMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Still Image"); break;
                    case 1: printf("Time Lapse"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 40:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Folder Name",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"FolderName",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Standard form"); break;
                    case 1: printf("Data form"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 41:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Color Mode",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"ColorMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Narural Color"); break;
                    case 1: printf("Black and White"); break;
                    case 2: printf("Vivid Color"); break;
                    case 3: printf("Solarization"); break;
                    case 4: printf("Adobe RGB"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 42:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Color Filter",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"ColorFilter",value);
                break;
            case 43:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"BW Filter",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"BWFilter",value);
                break;
            case 44:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Internal Flash",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"InternalFlash",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("did not fire"); break;
                    case 1: printf("fired"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 45:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s",i,"Brightness Value");
                else
                    printf("CS_%02d_%-21.21s",i,"BrightnessValue");
                if(PRINT_RAW_VALUES)
                    printf(" = %lu",value);
                fvalue = ((double)value/ 8.0) - 6.0;
                printf(" = %g APEX",fvalue);
                fvalue = pow(2.0,fvalue);
                print_startvalue();
                printf("%.3f foot lambert",fvalue);
                print_endvalue();
                break;
            case 46:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Spot Focus Point X",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"SpotFocusPoint X",value);
                break;
            case 47:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Spot Focus Point Y",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"SpotFocusPoint Y",value);
                break;
            case 48:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"White Focus Zone",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"WhiteFocusZone",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("no zone"); break;
                    case 1: printf("center zone (horizontal)"); break;
                    case 2: printf("center zone (vertical)"); break;
                    case 3: printf("left zone"); break;
                    case 4: printf("right zone"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 49:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Focus Mode",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"FocusMode",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Auto"); break;
                    case 1: printf("Manual"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 50:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Focus Area",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"FocusArea",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Wide"); break;
                    case 1: printf("Spot"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 51:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"DEC Position",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"DECPosition",value);
                print_startvalue();
                switch(value)
                {
                    case 0: printf("Exposure"); break;
                    case 1: printf("Contrast"); break;
                    case 2: printf("Saturation"); break;
                    case 3: printf("Filter"); break;
                    default: printred("undefined"); break;
                }
                print_endvalue();
                break;
            case 52:
                if(model == MINOLTA_DIMAGE7Hi)
                {
                    if(PRINT_SECTION)
                        printf("CS_%02d:%-21.21s = %lu",i,"Color Profile",value);
                    else
                        printf("CS_%02d_%-21.21s = %lu",i,"ColorProfile",value);
                    print_startvalue();
                    switch(value)
                    {
                        case 0: printf("Not Embedded"); break;
                        case 1: printf("Embedded"); break;
                        default: printred("undefined"); break;
                    }
                    print_endvalue();
                }
                else
                {
                    if(PRINT_SECTION)
                        printf("CS_%02d:%-21.21s = %lu",i,"Data Imprint",value);
                    else
                        printf("CS_%02d_%-21.21s = %lu",i,"DataImprint",value);
                    print_startvalue();
                    switch(value)
                    {
                        case 0: printf("None"); break;
                        case 1: printf("YYYY/MM/DD"); break;
                        case 2: printf("MM/DD/HR.MIN"); break;
                        case 3: printf("Text"); break;
                        case 4: printf("Text + Id#"); break;
                        default: printred("undefined"); break;
                    }
                    print_endvalue();
                }
                break;
            case 53:
                if(model == MINOLTA_DIMAGE7Hi)
                {
                    if(PRINT_SECTION)
                        printf("CS_%02d:%-21.21s = %lu",i,"Data Imprint",value);
                    else
                        printf("CS_%02d_%-21.21s = %lu",i,"DataImprint",value);
                    print_startvalue();
                    switch(value)
                    {
                        case 0: printf("None"); break;
                        case 1: printf("YYYY/MM/DD"); break;
                        case 2: printf("MM/DD/HR.MIN"); break;
                        case 3: printf("Text"); break;
                        case 4: printf("Text + Id#"); break;
                        default: printred("undefined"); break;
                    }
                    print_endvalue();
                }
                else
                {
                    if(PRINT_SECTION)
                        printf("CS_%02d:%-21.21s = %lu",i,"Unknown",value);
                    else
                        printf("CS_%02d_%-21.21s = %lu",i,"Unknown",value);
                }
                break;
            case 54:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Unknown",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Unknown",value);
                break;
            case 55:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Unknown",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Unknown",value);
                break;
            case 56:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Unknown",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Unknown",value);
                break;
            case 57:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"Unknown",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"Unknown",value);
                break;
            default:
                if(PRINT_SECTION)
                    printf("CS_%02d:%-21.21s = %lu",i,"undefined",value);
                else
                    printf("CS_%02d_%-21.21s = %lu",i,"undefined",value);
                break;
        }
        putchar('\n');
    }
}
