/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: print.c,v 1.28 2003/08/02 02:11:53 alex Exp $";
#endif

/* This file contains the primary "print" routines for TIFF/EXIF      */
/* data, file offsets, etc.                                           */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "tags.h"


/* Print the name, tagnumber and (where applicable) "version" of JPEG */
/* and TIFF header markers.                                           */
/* Return non-zero if the passed marker is not a proper header marker */
/* or 0 if it is,                                                     */

int
print_header(unsigned short marker,unsigned long section_id)
{
    char *markerstr;
    char *magicstr = "";
    int status = -1;

    switch(marker)
    {
    case JPEG_SOI:
        markerstr = "JPEG SOI";
        magicstr="\n";
        status = 0;
        break;
    case TIFF_INTEL:
        markerstr = "TIFF(II)";
        /* Assuming read_imagemarker() succeeded                      */
        magicstr = "0x42";
        status = 0;
        break;
    case TIFF_MOTOROLA:
        markerstr = "TIFF(MM)";
        /* Assuming read_imagemarker() succeeded                      */
        magicstr = "0x42";
        status = 0;
        break;
    default:
        markerstr = "UNKNOWN";
        magicstr="\n";
        break;
    }
    if(Print_options & section_id)
        printf("%s %#06x %s",markerstr,marker,magicstr);
    return(status);
}

void
print_imagetype(unsigned short marker)
{
    if((PRINT_LIST))
        printf("FileFormat = ");
    else
        printf("File Format = ");
    switch(marker)
    {
    case JPEG_SOI:
        printf("JPEG\n");
        break;
    case TIFF_INTEL:
    case TIFF_MOTOROLA:
        printf("TIFF\n");
        break;
    default:
        printf("UNKNOWN\n");
        break;
    }
}

/* Print an address in the file (offset from beginning of file), in   */
/* hex, decimal, or both, offset from left margin by 'indent' spaces. */
/* "Addresses" are usually prefixed by an "@", but may be prefixed by */
/* a string passed as argument (this is normally used to pass a '-'   */
/* prefix to mark the end of a segment, or something to flag an       */
/* out-of-order segment).                                             */

/* If this routine is called with all PRINT_ADDRESS bits off, it     */
/* produces only the indent.                                          */

void
print_tag_address(unsigned long section_id,unsigned long address,int indent,char *prefix,int addrwidth)
{
    if(Print_options & section_id)
    {
        if(PRINT_INDENT_BEFORE_ADDRESS)
            putindent(indent);
        if(!prefix)
            prefix = "@";
        if(PRINT_BOTH_ADDRESS)
            printf("%s%#0*lx=%-*lu:  ",prefix,addrwidth+1,address,addrwidth,address);
        else if(PRINT_HEX_ADDRESS)
            printf("%s%#0*lx:  ",prefix,addrwidth+1,address);
        else if(PRINT_DEC_ADDRESS)
            printf("%s%-*lu:  ",prefix,addrwidth,address);
        if(PRINT_INDENT_AFTER_ADDRESS)
            putindent(indent);
    }
}

/* Print the option-selected items in a TIFF IFD entry, including the */
/* tag number, tag name, tag type, and offset or value. If the        */
/* "value" doesn't fit in 4 bytes, print_offset_value() will be       */
/* called and may defer printing to be handled later (by the caller's */
/* second pass) or may print the offset, preceded by an '@' sign if   */
/* appropriate. If the value does fit in the 4 bytes provided in the  */
/* entry, print it according to its TIFF type, intercepting tags      */
/* which require special handling of value.                           */

unsigned long
print_entry(FILE *inptr,unsigned short byteorder,int addrwidth,
                    int ifdtype,int ifdnum,int subifdnum,
                    struct ifd_entry *entry_ptr,
                    unsigned long fileoffset_base,int indent)
{
    int value_is_offset = 0;
    unsigned long endoffset = 0UL;

    value_is_offset = is_offset(entry_ptr);
    if(PRINT_ENTRY)
    {
        print_taginfo(entry_ptr,indent,ifdtype,ifdnum,subifdnum);
        if((entry_ptr->value_type == 0) || (entry_ptr->value_type > DOUBLE) ||
            ferror(inptr) || feof(inptr))
        {
            printf(" INVALID ENTRY (%lu)",entry_ptr->value);
            endoffset = ftell(inptr);
        }
        else if(entry_ptr->count == 0)
        {
            if(PRINT_VALUE)
                printf(" EMPTY (value=%#lx)",entry_ptr->value);
            endoffset = ftell(inptr);
        }
        else if(value_is_offset)
        {
            /* Offsets read from the file are printed relative to the */
            /* beginning of the file (with fileoffset_base added) so  */
            /* that the values, printed later, will be easy to find.  */
            if(PRINT_BOTH_OFFSET)
                printf("@%#lx=%lu",entry_ptr->value + fileoffset_base,
                        entry_ptr->value + fileoffset_base);
            else if(PRINT_HEX_OFFSET)
                printf("@%#lx",entry_ptr->value + fileoffset_base);
            else if(PRINT_DEC_OFFSET)
                printf("@%lu",entry_ptr->value + fileoffset_base);

            if(fileoffset_base && (PRINT_ENTRY_RELOFFSET))
                printf(" (%lu) ",entry_ptr->value);

            if((PRINT_VALUE) && !(PRINT_VALUE_AT_OFFSET))
            {
                if((PRINT_OFFSET) && (entry_ptr-> tag != EXIFTAG_MakerNote))
                    printf(" = ");
                endoffset = print_offset_value(inptr,byteorder,
                                        entry_ptr,fileoffset_base,
                                        ifdtype,indent,addrwidth,0);
                return(endoffset);
            }
            else if((PRINT_VALUE == 0))
            {
                /* Really done printing, but run through this just to */
                /* check ASCII entries for Make, Model, or Software,  */
                /* which must be recorded.                            */
                endoffset = print_offset_value(inptr,byteorder,
                                        entry_ptr,fileoffset_base,
                                        ifdtype,indent,addrwidth,0);
            }
        }
        else if(PRINT_VALUE)
        {
            /* Intercept some EXIF values which print_value() would   */
            /* not handle properly, possibly because they are ascii   */
            /* items tagged as UNDEFINED. A few items given as LONG   */
            /* type are actually offsets, and are printed as such     */
            /* here, with fileoffset_base added.                      */
            if((ifdtype == INTEROP_IFD) && (entry_ptr->tag == INTEROPTAG_Version))
                printf("%s\"%s\"",PREFIX,(char *)&entry_ptr->value);
            else
            {
                switch(entry_ptr->tag)
                {
                    case EXIFTAG_ExifVersion:
                    case EXIFTAG_FlashPixVersion:
                        /* ASCII                                      */
                        printf("%s\"%s\"",PREFIX,(char *)&entry_ptr->value);
                        break;
                    case EXIFTAG_ExifIFDPointer:
                    case EXIFTAG_GPSInfoIFDPointer:
                    case EXIFTAG_Interoperability:
                    case TIFFTAG_JPEGInterchangeFormat:
                        /* OFFSETS                                    */
                        if(PRINT_BOTH_OFFSET)
                            printf("@%#lx=%lu",entry_ptr->value + fileoffset_base,
                                    entry_ptr->value + fileoffset_base);
                        else if(PRINT_HEX_OFFSET)
                            printf("@%#lx",entry_ptr->value + fileoffset_base);
                        else if((PRINT_DEC_OFFSET) || is_a_long_offset(entry_ptr))
                            printf("@%lu",entry_ptr->value + fileoffset_base);

                        if(fileoffset_base && (PRINT_ENTRY_RELOFFSET))
                            printf(" (%lu)",entry_ptr->value);
                        break;
                    default:
                        /* VALUES                                     */
                        print_value(entry_ptr,PREFIX);
                        interpret_value(entry_ptr);
                        break;
                }
            }
        }
        (void)putchar('\n');
    }
    return(endoffset);
}

int
is_a_long_offset(struct ifd_entry *entry_ptr)
{
    int is_an_offset = 0;

    if(entry_ptr)
    {
        switch(entry_ptr->tag)
        {
            case EXIFTAG_ExifIFDPointer:
            case EXIFTAG_GPSInfoIFDPointer:
            case EXIFTAG_Interoperability:
            case TIFFTAG_JPEGInterchangeFormat:
                ++is_an_offset;
                break;
            default:
                break;
        }
    }
    return(is_an_offset);
}

/* Repeat the tag identifier for values printed at an offset. This is */
/* called only from the print_offset_value() routine when printing    */
/* offset values at the offset (rather than "inline"), and repeats    */
/* only one of the identifiers (preferably tagname) followed by the   */
/* value separator (if values are printed at all). If no tag          */
/* identifier is enabled in Print_options, the tag name will be       */
/* printed, as in the entry line printed by print_taginfo().          */

void
print_tagid(struct ifd_entry *entry_ptr,int indent,int ifdtype)
{
    char *nameoftag;
    char *nameof_value_type;

    if(entry_ptr && (PRINT_ENTRY))
    {
        extraindent(indent);
        if(PRINT_TAGNAME || !(PRINT_TAGNO))
        {
            switch(ifdtype)
            {
                case GPS_IFD:
                    nameoftag = gps_tagname(entry_ptr->tag);
                    break;
                case INTEROP_IFD:
                    nameoftag = interop_tagname(entry_ptr->tag);
                    break;
                case EXIF_IFD:
                case TIFF_IFD:
                case TIFF_SUBIFD:
                default:
                    nameoftag = tagname(entry_ptr->tag);
                    break;
            }
            /* should line up next item with print_taginfo() output.  */
            printf("%-*.*s",IDWIDTH,IDWIDTH,nameoftag);
        }
        else if(PRINT_HEX_TAGNO)
            printf("<%#06x>",entry_ptr->tag & 0xffff);
        else
            printf("<%5u>",entry_ptr->tag & 0xffff);
        if(PRINT_VALUE)
            printf(" = ");
    }
}


/* Print the part of an IFD entry describing the entry tag, including */
/* it's tag number, name and type. Only the items enabled in          */
/* "Print_options" are printed, except that if the value is to be     */
/* printed but neither tagname or tag number is enabled, print the    */
/* tag name anyway (it doesn't make much sense to print the value     */
/* without some sort of identifer).                                   */

/* In LIST mode (when not printing SECTION info) prefix the tag name  */
/* with a section identifier, so that the resulting names may be      */
/* recognized by patterns or regular expressions applied to the       */
/* output (e.g. 'egrep') in post-processing.                          */

void
print_taginfo(struct ifd_entry *entry_ptr,int indent,int ifdtype,
                int ifdnum,int subifdnum)
{
    char prefixbuf[16];
    char *nameoftag;
    char *nameof_value_type;
    char *tagprefix = "";
    int tagwidth = TAGWIDTH;

    if(entry_ptr && (PRINT_ENTRY))
    {
        extraindent(indent);
        if(PRINT_BOTH_TAGNO)
            printf("<%#06x=%5u> ",entry_ptr->tag & 0xffff,entry_ptr->tag & 0xffff);
        else if(PRINT_DEC_TAGNO)
            printf("<%5u> ",entry_ptr->tag & 0xffff);
        else if(PRINT_HEX_TAGNO)
            printf("<%#06x> ",entry_ptr->tag & 0xffff);
        if((PRINT_TAGNAME) || ((PRINT_VALUE) && !(PRINT_TAGNO)))
        {
            switch(ifdtype)
            {
                case GPS_IFD:
                    if((PRINT_LIST))
                    {
                        tagprefix = "GPS_";
                        ++tagwidth;
                    }
                    nameoftag = gps_tagname(entry_ptr->tag);
                    break;
                case INTEROP_IFD:
                    if((PRINT_LIST))
                        tagprefix = "Exif_";
                    nameoftag = interop_tagname(entry_ptr->tag);
                    break;
                case EXIF_IFD:
                    if((PRINT_LIST))
                        tagprefix = "Exif_";
                    nameoftag = tagname(entry_ptr->tag);
                    break;
                case TIFF_IFD:
                    if(((PRINT_LIST)) /* && (ifdnum > 0) */)
                    {
                        sprintf(prefixbuf,"Ifd%d_",ifdnum);
                        tagprefix = prefixbuf;
                    }
                    nameoftag = tagname(entry_ptr->tag);
                    break;
                case TIFF_SUBIFD:
                    if(((PRINT_LIST)) /* && (ifdnum > 0) */)
                    {
                        sprintf(prefixbuf,"SubIfd%d%d_",ifdnum,subifdnum);
                        tagprefix = prefixbuf;
                        tagwidth -= 4;
                    }
                    nameoftag = tagname(entry_ptr->tag);
                    break;
                default:
                    nameoftag = tagname(entry_ptr->tag);
                    break;
            }
            /* Print everything same width, so that following items   */
            /* line up.  tagprefix should be 5 chars wide.            */
            if((PRINT_SECTION == 0) && (*tagprefix == '\0'))
                tagwidth += PREFIXLEN;
            printf("%s%-*.*s",tagprefix,tagwidth,tagwidth,nameoftag);
        }
        if(PRINT_TYPE)
        {
            nameof_value_type = value_type_name(entry_ptr->value_type);
            printf(" [%-2u=%-9.9s %5lu] ",entry_ptr->value_type,
                    nameof_value_type,entry_ptr->count);
        }
        if(PRINT_VALUE)
            printf(" = "); 
    }
}

/* Print a "value" from an ifd structure entry, according to the TIFF */
/* type recorded in the entry. Print in hex or decimal (or both)      */
/* according to the global Print_options variable.                    */

void
print_value(struct ifd_entry *entry_ptr,char *prefix)
{
    unsigned long byte;
    unsigned short ushort_val;
    int i;

    if(prefix)
        printf("%s",prefix);
    switch(entry_ptr->value_type)
    {
        case SBYTE:
                    byte = entry_ptr->value;
                    for(i = 0; i < entry_ptr->count; ++i)
                    {
                        if(PRINT_BOTH_VALUE)
                            printf("%#x=%d",(unsigned int)((char)byte & 0xff),
                                        (int)((char)byte & 0xff));
                        else if(PRINT_HEX_VALUE)
                            printf("%#x",(unsigned int)((char)byte & 0xff));
                        else if(PRINT_DEC_VALUE)
                            printf("%d",(int)((char)byte & 0xff));
                        byte >>= 8;
                    }
                    break;
        case SHORT:
                    ushort_val = entry_ptr->value & 0xffff;
                    if(PRINT_BOTH_VALUE)
                        printf("%#x=%u",ushort_val,ushort_val);
                    else if(PRINT_HEX_VALUE)
                        printf("%#x",ushort_val);
                    else if(PRINT_DEC_VALUE)
                        printf("%u",ushort_val);
                    if(entry_ptr->count > 1)
                    {
                        (void)putchar(',');
                        ushort_val = (entry_ptr->value & 0xffff0000) >> 16;
                        if(PRINT_BOTH_VALUE)
                            printf("%#x=%u",ushort_val,ushort_val);
                        else if(PRINT_HEX_VALUE)
                            printf("%#x",ushort_val);
                        else if(PRINT_DEC_VALUE)
                            printf("%u",ushort_val);
                    }
                    break;
        case SSHORT:
                    ushort_val = entry_ptr->value & 0xffff;
                    if(PRINT_BOTH_VALUE)
                        printf("%#x=%d",(unsigned int)ushort_val,
                                    (int)ushort_val);
                    else if(PRINT_HEX_VALUE)
                        printf("%#x",(unsigned int)ushort_val);
                    else if(PRINT_DEC_VALUE)
                        printf("%d",(int)ushort_val);
                    if(entry_ptr->count > 1)
                    {
                        (void)putchar(',');
                        ushort_val = (entry_ptr->value & 0xffff0000) >> 16;
                        if(PRINT_BOTH_VALUE)
                            printf("%#x=%d",(unsigned int)ushort_val,
                                        (int)ushort_val);
                        else if(PRINT_HEX_VALUE)
                            printf("%#x",(unsigned int)ushort_val);
                        else if(PRINT_DEC_VALUE)
                            printf("%d",(int)ushort_val);
                    }
                    break;
        case LONG:
                    if(PRINT_BOTH_VALUE)
                        printf("0x%08lx=%lu",entry_ptr->value,
                                    entry_ptr->value);
                    else if(PRINT_HEX_VALUE)
                        printf("0x%08lx",entry_ptr->value);
                    else if(PRINT_DEC_VALUE)
                        printf("%lu",entry_ptr->value);
                    break;
        case SLONG:
                    if(PRINT_BOTH_VALUE)
                        printf("%#lx=%ld",(unsigned long)entry_ptr->value,
                                    (long)entry_ptr->value);
                    else if(PRINT_HEX_VALUE)
                        printf("%#lx",(unsigned long)entry_ptr->value);
                    else if(PRINT_DEC_VALUE)
                        printf("%ld",(long)entry_ptr->value);
                    break;
        case ASCII:
                    printf("\"%*.*s\"",(int)entry_ptr->count,
                            (int)entry_ptr->count,(char *)&entry_ptr->value);
                    break;
        case FLOAT:
                    printf("%g",(float)entry_ptr->value);
                    break;
        case DOUBLE: /* double is not possible in a direct ifd entry  */
                    printf("INVALID");
                    break;
        case UNDEFINED: /* fall through; handle these as BYTE         */
        case BYTE:
        default:
                    byte = entry_ptr->value;
                    for(i = 0; i < entry_ptr->count; ++i)
                    {
                        if(i)
                            (void)putchar(',');
                        if(PRINT_BOTH_VALUE)
                            printf("%#lx=%lu",byte & 0xff,byte & 0xff);
                        else if(PRINT_HEX_VALUE)
                            printf("%#lx",byte & 0xff);
                        else if(PRINT_DEC_VALUE)
                            printf("%lu",byte & 0xff);
                        byte >>= 8;
                    }
                    /* UNDEFINED stuff may be ascii                   */
                    byte = entry_ptr->value;
                    if((entry_ptr->value_type == UNDEFINED) &&
                        (entry_ptr->count == 4) &&
                            isprint(byte & 0xff))   
                            
                    {
                        printf(" = \"");
                        for(i = 0; i < entry_ptr->count; ++i)
                        {
                            if(isprint(byte & 0xff))
                                (void)putchar(byte & 0xff);
                            else
                                (void)putchar('.');
                            byte >>= 8;
                        }
                        (void)putchar('"');
                    }
                    break;
    }
}

/* Goto the offset given by an IFD entry whose value couldn't be      */
/* squeezed into the 4 bytes of the entry itself, and read the value  */
/* found at that offset, printing according to the TIFF type found in */
/* the entry. UNDEFINED entry types must/may be special-cased, else   */
/* their presence and size will be noted and the data skipped or      */
/* hexdumped.                                                         */

/* Returns the file pointer offset at the end of processing, which    */
/* may be 0 if the seek or read fails, or if the entry does not have  */
/* offset data, or the data is outside the current IFD (e.g.          */
/* JPEGTables).                                                       */

unsigned long
print_offset_value(FILE *inptr,unsigned short byteorder,
                struct ifd_entry *entry_ptr,unsigned long fileoffset_base,
                int ifdtype,int indent,int addrwidth,int at_offset)
{
    unsigned long endoffset = 0L;
    unsigned long value_offset,dumplength;
    unsigned short marker;
    char *nameoftag;
    char *makename,*modelname,*swname;

    if(inptr && entry_ptr)
    {
        value_offset = fileoffset_base + entry_ptr->value;
        endoffset = value_offset + 
            (value_type_size(entry_ptr->value_type) * entry_ptr->count);

        if(PRINT_VALUE)
        {
            switch(entry_ptr->value_type)
            {
                case UNDEFINED: 
                    switch(entry_ptr->tag)
                    {
                        case EXIFTAG_CFAPattern:
                            print_cfapattern(inptr,entry_ptr->count,
                                                byteorder,value_offset);
                            break;
                        case EXIFTAG_UserComment:
                            print_user_comment(inptr,entry_ptr->count,
                                                value_offset);
                            break;
                        case TIFFTAG_JPEGTables:
                            nameoftag = tagname(entry_ptr->tag);
                            if(PRINT_SEGMENT)
                            {
                                printf("Start of %s length %lu\n",nameoftag,
                                                            entry_ptr->count);
                            }
                            else if(!(PRINT_OFFSET))
                            {
                                if(PRINT_BOTH_VALUE)
                                    printf("@%#lx=%lu",value_offset,value_offset);
                                else if(PRINT_HEX_VALUE)
                                    printf("@%#lx",value_offset);
                                else if(PRINT_DEC_VALUE)
                                    printf("@%lu",value_offset);
                            }
                            marker = read_imagemarker(inptr,value_offset);
                            (void)process_jpeg_segments(inptr,value_offset,
                                                        marker,entry_ptr->count,
                                                        indent+2,addrwidth);
                            if((PRINT_SEGMENT))
                            {
                                print_tag_address(ENTRY,value_offset + entry_ptr->count - 1,
                                                    indent,"-",addrwidth);
                                extraindent(SUBINDENT);
                                printf("End of %s",nameoftag);
                            }
                            break;
                        case EXIFTAG_MakerNote:
                            if(!at_offset && !(PRINT_OFFSET))
                            {
                                if(PRINT_BOTH_VALUE)
                                    printf("@%#lx=%lu",value_offset,value_offset);
                                else if(PRINT_HEX_VALUE)
                                    printf("@%#lx",value_offset);
                                else if((PRINT_DEC_VALUE) || is_a_long_offset(entry_ptr))
                                    printf("@%lu",value_offset);
                            }
                            break;
                        /* TIFFEP_CFAPATTERN is not quite the same as */
                        /* EXIF_CFAPattern, need CFARepeatPatternDim, */
                        /* and some way of remembering it. Ick...     */
                        case TIFFEPTAG_CFAPattern:
                        case TIFFEPTAG_SpatialFrequencyResponse:
                        case EXIFTAG_SpatialFrequencyResponse:
                        case EXIFTAG_SpectralSensitivity:
                        case EXIFTAG_OECF:
                            /* fall through, until some routines are  */
                            /* written for these.                     */
                        default:
                            printf("UNDEFINED, count %lu",entry_ptr->count);
                            if(Max_undefined == 0L)
                                printf(" (not dumped, use -U)");
                            else
                            {
                                if((Max_undefined == DUMPALL) ||
                                        (Max_undefined > entry_ptr->count))
                                    dumplength = entry_ptr->count;
                                else 
                                    dumplength = Max_undefined;
                                putchar('\n');
                                hexdump(inptr,value_offset,entry_ptr->count,
                                            dumplength,12,indent,SUBINDENT,
                                            addrwidth);
                            }
                            /* make certain we're at the end          */
                            fseek(inptr,value_offset + entry_ptr->count,0);
                            break;
                    }
                    break;
                case BYTE:
                    print_ubytes(inptr,entry_ptr->count,value_offset);
                    break;
                case SBYTE:
                    print_sbytes(inptr,entry_ptr->count,value_offset);
                    break;
                case ASCII:
                    print_ascii(inptr,entry_ptr->count,value_offset); break;
                    break;
                case SHORT:
                    print_ushort(inptr,entry_ptr->count,byteorder,value_offset);
                    break;
                case SSHORT:
                    print_sshort(inptr,entry_ptr->count,byteorder,value_offset);
                    break;
                case LONG:
                    print_ulong(inptr,entry_ptr->count,byteorder,value_offset);
                    break;
                case SLONG:
                    print_slong(inptr,entry_ptr->count,byteorder,value_offset);
                    break;
                case RATIONAL:
                    print_urational(inptr,entry_ptr->count,byteorder,value_offset);
                    interpret_rational(inptr,byteorder,entry_ptr->tag,value_offset);
                    break;
                case SRATIONAL:
                    print_srational(inptr,entry_ptr->count,byteorder,value_offset);
                    interpret_rational(inptr,byteorder,entry_ptr->tag,value_offset);
                    break;
                case FLOAT:
                    print_float(inptr,entry_ptr->count,byteorder,value_offset);
                    break;
                case DOUBLE:
                    print_double(inptr,entry_ptr->count,byteorder,value_offset);
                    break;
                default:
                    printf(" INVALID TYPE %#x",entry_ptr->value_type);
                    break;
            }
            (void)putchar('\n');
        }
        if(entry_ptr->value_type == ASCII)
        {
            /* These items may have been read and printed above, but  */
            /* must be read again and saved for possible use when     */
            /* processing MakerNotes. This is slightly inefficient,   */
            /* and if no MakerNote is processed it is a complete      */
            /* waste; it is the price we pay for a single-pass        */
            /* "print-and-forget" process. These are the only items   */
            /* that the process "remembers".                          */

            /* The names are not recorded if they already exist, in   */
            /* order to support the (undocumented) options to set     */
            /* them specifically 'by hand'                            */
            switch(entry_ptr->tag)
            {
                case 0x10F:
                    makename = (char *)read_bytes(inptr,entry_ptr->count,
                                                    value_offset);
                    if(Make_name == NULL)
                        Make_name = strdup(makename);
                    break;
                case 0x110:
                    modelname = (char *)read_bytes(inptr,entry_ptr->count,
                                                    value_offset);
                    if(Model_name == NULL)
                        Model_name = strdup(modelname);
                    break;
                case 0x131:
                    swname = (char *)read_bytes(inptr,entry_ptr->count,
                                                    value_offset);
                    if(Software_name == NULL)
                        Software_name = strdup(swname);
                    break;
            }

        }
    }
    else
    {
        printf(" invalid call to print_offset_value() ");
        if(inptr == (FILE *)0)
            printf(" - no input file pointer ");
        if(entry_ptr == (struct ifd_entry *)0)
            printf(" - null entry pointer");
    }
    return(endoffset);
}

/* Read 'count' unsigned bytes starting at 'offset' and print their   */
/* values in hex. If there are a lot, just print the first few, an    */
/* elipsis to admit that something is left out, then print the count  */
/* and the last value and it's offset.                                */

void
print_ubytes(FILE *inptr,unsigned long count,unsigned long offset)
{
    int i,j;
    unsigned short byte;

    if(count)
    {
        byte = read_ubyte(inptr,offset);
        if(feof(inptr) || ferror(inptr))
            return;
        printf("%02x",(unsigned int)byte & 0xff);
    }
    if(PRINT_MULTIVAL_ALL)
    {
        for(i = 1; i < count; ++i,offset += 2L)
        {
            byte = read_ubyte(inptr,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%02x",(unsigned int)byte & 0xff);
        }
    }
    else
    {
#define MAX 10UL
        j = count > MAX + 2 ? MAX : count;
        for(i = 1; i < j; ++i)
        {
            byte = read_ubyte(inptr,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%02x",(unsigned int)byte & 0xff);
        }
        if((count > j) && (ferror(inptr) == 0))
        {
            printf(" ... ");
            offset += (unsigned long)(count - 1);
            byte = read_ubyte(inptr,offset);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%02x (%lu)",(unsigned int)byte & 0xff,count);
            if(PRINT_ADDRESS)
                printf(" -> @%lu",offset);
        }
#undef MAX
    }
}

/* Read 'count' signed bytes starting at 'offset' and print their     */
/* values. If there are a lot, just print the first few, an elipsis   */
/* to admit that something is left out, then print the last value and */
/* it's offset.                                                       */

void
print_sbytes(FILE *inptr,unsigned long count,unsigned long offset)
{
    int i,j;
    short byte;

    if(count)
    {
        byte = read_ubyte(inptr,offset);
        if(feof(inptr) || ferror(inptr))
            return;
        printf("%d",byte);
    }
    if(PRINT_MULTIVAL_ALL)
    {
        for(i = 1; i < count; ++i,offset += 2L)
        {
            byte = read_ubyte(inptr,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%d",byte);
        }
    }
    else
    {
#define MAX 10UL
        j = count > MAX + 2 ? MAX : count;
        for(i = 1; i < j; ++i)
        {
            byte = read_ubyte(inptr,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%d",byte);
        }
        if((count > j) && (ferror(inptr) == 0))
        {
            printf(" ... ");
            offset += (unsigned long)(count - 1);
            byte = read_ubyte(inptr,offset);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%d (%lu)",byte,count);
            if(PRINT_ADDRESS)
                printf(" -> @%lu",offset);
        }
#undef MAX
    }
}

/* Read 'count' unsigned short integers starting at 'offset' and      */
/* print their values. If there are a lot, just print the first few,  */
/* an elipsis to admit that something is left out, then print the     */
/* last value and it's offset.                                        */

void
print_ushort(FILE *inptr,unsigned long count,unsigned short byteorder,
                unsigned long offset)
{
    int i,j;
    unsigned short value;

    if(count)
    {
        value = read_ushort(inptr,byteorder,offset);
        if(feof(inptr) || ferror(inptr))
            return;
        printf("%u",value);
    }
    if(PRINT_MULTIVAL_ALL)
    {
        for(i = 1; i < count; ++i,offset += 2L)
        {
            value = read_ushort(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%u",value);
        }
    }
    else
    {
#define MAX 5UL
        j = count > MAX + 2 ? MAX : count;
        for(i = 1; i < j; ++i,offset += 2L)
        {
            value = read_ushort(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%u",value);
        }
        if((count > j) && (ferror(inptr) == 0))
        {
            printf(" ... ");
            offset += (unsigned long)(count - j) * 2L;
            value = read_ushort(inptr,byteorder,offset);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%u (%lu)",value,count);
            if(PRINT_ADDRESS)
                printf(" -> @%lu",offset);
        }
#undef MAX
    }
}

/* Read 'count' signed short integers starting at 'offset' and print  */
/* their values. If there are a lot, just print the first 5, an       */
/* elipsis to admit that something is left out, then print the last   */
/* value and it's offset.                                             */

void
print_sshort(FILE *inptr,unsigned long count,unsigned short byteorder,
                unsigned long offset)
{
    int i,j;
    short value;

    if(count)
    {
        value = (short)read_ushort(inptr,byteorder,offset);
        if(feof(inptr) || ferror(inptr))
            return;
        printf("%d",value);
    }
    if(PRINT_MULTIVAL_ALL)
    {
        for(i = 1; i < count; ++i,offset += 2L)
        {
            value = (short)read_ushort(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%d",value);
        }
    }
    else
    {
#define MAX 5UL
        j = count > MAX + 2 ? MAX : count;
        for(i = 1; i < j; ++i,offset += 2L)
        {
            value = (short)read_ushort(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%d",value);
        }
        if(count > j)
        {
            printf(" ... ");
            offset += (unsigned long)(count - j) * 2L;
            value = (short)read_ushort(inptr,byteorder,offset);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%d (%lu)",value,count);
            if(PRINT_ADDRESS)
                printf(" -> @%lu",offset);
        }
#undef MAX
    }
}

/* Read 'count' unsigned long integers starting at 'offset' and print */
/* their values. If there are a lot, just print the first 3, an       */
/* elipsis to admit that something is left out, then print the last   */
/* value and it's offset.                                             */

void
print_ulong(FILE *inptr,unsigned long count,unsigned short byteorder,
                unsigned long offset)
{
    int i,j;
    unsigned long value;

    if(count)
    {
        value = read_ulong(inptr,byteorder,offset);
        if(feof(inptr) || ferror(inptr))
            return;
        printf("%lu",value);
    }
    if(PRINT_MULTIVAL_ALL)
    {
        for(i = 1; i < count; ++i,offset += 4L)
        {
            value = read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%lu",value);
        }
    }
    else
    {
#define MAX 3UL
        j = count > MAX + 2 ? MAX : count;
        for(i = 1; i < j; ++i,offset += 4L)
        {
            value = read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%lu",value);
        }
        if(count > j)
        {
            printf(" ... ");
            offset += (unsigned long)(count - j) * 4L;
            value = read_ulong(inptr,byteorder,offset);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%lu (%lu)",value,count);
            if(PRINT_ADDRESS)
                printf(" -> @%lu",offset);
        }
#undef MAX
    }
}

/* Read 'count' signed long integers starting at 'offset' and print   */
/* their values. If there are a lot, just print the first 3, an       */
/* elipsis to admit that something is left out, then print the last   */
/* value and it's offset.                                             */

void
print_slong(FILE *inptr,unsigned long count,unsigned short byteorder,
                unsigned long offset)
{
    int i,j;
    long value;

    if(count)
    {
        value = (long)read_ulong(inptr,byteorder,offset);
        if(feof(inptr) || ferror(inptr))
            return;
        printf("%ld",value);
    }
    if(PRINT_MULTIVAL_ALL)
    {
        for(i = 1; i < count; ++i,offset += 4L)
        {
            value = (long)read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%ld",value);
        }
    }
    else
    {
#define MAX 3UL
        j = count > MAX + 2 ? MAX : count;
        for(i = 1; i < j; ++i)
        {
            value = (long)read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%ld",value);
        }
        if(count > j)
        {
            printf(" ... ");
            offset += (unsigned long)(count - j) * 4L;
            value = (long)read_ulong(inptr,byteorder,offset);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%ld (%lu)",value,count);
            if(PRINT_ADDRESS)
                printf(" -> @%lu",offset);
        }
#undef MAX
    }
}

/* Read 'count' unsigned rational values starting at 'offset' and     */
/* print their values. If there are a lot, just print the first 3, an */
/* elipsis to admit that something is left out, then print the last   */
/* value and it's offset.                                             */

void
print_urational(FILE *inptr,unsigned long count,unsigned short byteorder,
                unsigned long offset)
{
    int i,j;
    unsigned long num,denom;

    if(count)
    {
        num = read_ulong(inptr,byteorder,offset);
        if(feof(inptr) || ferror(inptr))
            return;
        denom = read_ulong(inptr,byteorder,HERE);
        if(feof(inptr) || ferror(inptr))
            return;
        if(PRINT_RAW_VALUES)
            printf("%lu/%lu = ",num,denom);
        if(num == 0UL)
            printf("0");
        else
            printf("%g",(double)((double)num/(double)denom));
    }
    if(PRINT_MULTIVAL_ALL)
    {
        for(i = 1; i < count; ++i)
        {
            num = read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            denom = read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            if(PRINT_RAW_VALUES)
                printf("%lu/%lu = ",num,denom);
            if(num == 0UL)
                printf(",0");
            else
                printf(",%g",(double)((double)num/(double)denom));
        }
    }
    else
    {
#define MAX 3UL
        j = count > MAX + 2 ? MAX : count;
        for(i = 1; i < j; ++i)
        {
            num = read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            denom = read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            if(PRINT_RAW_VALUES)
                printf("%lu/%lu = ",num,denom);
            if(num == 0UL)
                printf(",0");
            else
                printf(",%g",(double)((double)num/(double)denom));
        }
        if(count > j)
        {
            printf(" ... ");
            offset += (unsigned long)(count - j) * 8UL;
            fseek(inptr,offset,0);
            num = read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            denom = read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            if(PRINT_RAW_VALUES)
                printf("%lu/%lu = ",num,denom);
            else
            if(num == 0UL)
                printf(",0 (%lu)",count);
            else
                printf(",%g (%lu)",(double)((double)num/(double)denom),count);
            if(PRINT_ADDRESS)
                printf(" -> @%lu",offset);
        }
#undef MAX
    }
}

void
print_srational(FILE *inptr,unsigned long count,unsigned short byteorder,
                unsigned long offset)
{
    int i,j;
    long num,denom;

    if(count)
    {
        num = (long)read_ulong(inptr,byteorder,offset);
        if(feof(inptr) || ferror(inptr))
            return;
        denom = (long)read_ulong(inptr,byteorder,HERE);
        if(feof(inptr) || ferror(inptr))
            return;
        if(PRINT_RAW_VALUES)
            printf("%lu/%lu = ",num,denom);
        if(num == 0UL)
            printf("0");
        else
            printf("%g",(double)((double)num/(double)denom));
    }
    if(PRINT_MULTIVAL_ALL)
    {
        for(i = 1; i < count; ++i)
        {
            num = (long)read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            denom = (long)read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            if(PRINT_RAW_VALUES)
                printf("%lu/%lu = ",num,denom);
            if(num == 0UL)
                printf(",0");
            else
                printf(",%g",(double)((double)num/(double)denom));
        }
    }
    else
    {
#define MAX 3UL
        j = count > MAX + 2 ? MAX : count;
        for(i = 1; i < j; ++i)
        {
            num = (long)read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            denom = (long)read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            if(PRINT_RAW_VALUES)
                printf("%lu/%lu = ",num,denom);
            if(num == 0UL)
                printf(",0");
            else
                printf(",%g",(double)((double)num/(double)denom));
        }
        if(count > j)
        {
            printf(" ... ");
            offset += (unsigned long)(count - j) * 8UL;
            fseek(inptr,offset,0);
            num = (long)read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            denom = (long)read_ulong(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            if(PRINT_RAW_VALUES)
                printf("%lu/%lu = ",num,denom);
            if(num == 0UL)
                printf(",0 (%lu)",count);
            else
                printf(",%g (%lu)",(double)((double)num/(double)denom),count);
            if(PRINT_ADDRESS)
                printf(" -> @%lu",offset);
        }
#undef MAX
    }
}

void
print_float(FILE *inptr,unsigned long count,unsigned short byteorder,
                unsigned long offset)
{
    int i,j;
    double data;

    if(count)
    {
        data = read_float(inptr,byteorder,offset);
        if(feof(inptr) || ferror(inptr))
            return;
        printf("%g",data);
    }
    if(PRINT_MULTIVAL_ALL)
    {
        for(i = 1; i < count; ++i)
        {
            data = read_float(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%g",data);
        }
    }
    else
    {
#define MAX 3L
        j = count > MAX + 2 ? MAX : count;
        for(i = 1; i < j; ++i)
        {
            data = read_float(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%g",data);
        }
        if(count > j)
        {
            printf(" ... ");
            offset += (unsigned long)(count - j) * 4L;
            data = read_float(inptr,byteorder,offset);
            printf(",%g (%lu)",data,count);
            if(PRINT_ADDRESS)
                printf(" -> @%lu",offset);
        }
#undef MAX
    }
}


void
print_double(FILE *inptr,unsigned long count,unsigned short byteorder,
                unsigned long offset)
{
    int i,j;
    double data;

    if(count)
    {
        data = read_double(inptr,byteorder,offset);
        if(feof(inptr) || ferror(inptr))
            return;
        printf("%g",data);
    }
    if(PRINT_MULTIVAL_ALL)
    {
        for(i = 1; i < count; ++i)
        {
            data = read_double(inptr,byteorder,HERE);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%g",data);
        }
    }
    else
    {
#define MAX 3L
        j = count > MAX + 2 ? MAX : count;
        for(i = 1; i < j; ++i)
        {
            data = read_double(inptr,byteorder,HERE);
            printf(",%g",data);
        }
        if(count > j)
        {
            printf(" ... ");
            offset += (unsigned long)(count - j) * 8UL;
            data = read_double(inptr,byteorder,offset);
            if(feof(inptr) || ferror(inptr))
                return;
            printf(",%g (%lu)",data,count);
            if(PRINT_ADDRESS)
                printf(" -> @%lu",offset);
        }
#undef MAX
    }
}

/* Print 'count' bytes at 'offset' as an ascii string, expanding      */
/* nulls and unprintable characters.                                  */

int
print_ascii(FILE *inptr,unsigned long count,unsigned long offset)
{
    int remaining;
    int nread;
    int ch,lastch;

    remaining = count;
    nread = 0;
    if((offset != HERE) && (fseek(inptr,offset,0) == -1))
        printf(" seek failed to offset %lu to print ascii\n",offset);
    {
        lastch = 'A';
        (void)putchar('"');
        while((remaining > 0) && !feof(inptr))
        {
            ch = fgetc(inptr);
            ++nread;
            --remaining;
            if(ch && isprint(ch))
            {
                if(isdigit(ch) && (lastch == '\0'))
                    printf("00");
                (void)putchar(ch);
            }
            else if(ch == '\0')
            {
                /* Normally honor the count; show everything.         */
                /* Optionally (-ea) quit at the first null, like a    */
                /* normal string. The latter is better for listing.   */
                if(PRINT_ASCII_IGNORE_LENGTH)
                    break;
                else
                    printf("\\0");
            }
            else
                printf("\\%03o",(unsigned int)(ch & 0xff));
            lastch = ch;
        }
        (void)putchar('"');
        if(ferror(inptr))
            clearerr(inptr);
    }
    return(nread);
}

/* Print the UserComment field from an EXIF segment, which is         */
/* expected at 'offset' in the file and should be 'count' bytes long. */
/* The field begins with an 8 byte "character code", followed by      */
/* 'count - 8' bytes of actual comment. The length is printed as      */
/* "%d+8" to give the user the usable length of the comment, but      */
/* remind that the 8 bytes is there. An "undefined" field is printed  */
/* as ascii, but checked to see if it is filled with nulls or spaces. */
/* Non-ascii comment types are simply hexdumped.                      */

void
print_user_comment(FILE *inptr,unsigned long count,unsigned long offset)
{
    char *data;
    int i;

    /* The first 8 bytes identify the character code                  */

    if(PRINT_ENTRY)
    {
        if(PRINT_SECTION)
            printf("count %lu+8: (CC=",count - 8);
        if((offset != HERE) && (fseek(inptr,offset,0) == -1))
            printf("seek failed to offset %lu to process user comment)\n",offset);
        else if(data = (char *)read_bytes(inptr,8UL,HERE))
        {
            if(offset == HERE)
                offset = ftell(inptr);  /* need to return after check     */
            if((data[0] == 0) || (strncmp(data,"ASCII",5) == 0))
            {
                if(PRINT_SECTION)
                {
                    if(data[0] == 0)
                        printf("undef");
                    else
                    {
                        for(i = 0; i < 8; ++i)
                        {
                            if(data[i] && isascii(data[i]))
                                (void)putchar(data[i]);
                        }
                    }
                    (void)putchar(')');
                }
                i = 8;
                if(PRINT_SECTION)
                {
                    /* check for all nulls                                   */
                    for( ; i < count; ++i)
                    {
                        if(fgetc(inptr) != '\0')
                            break;
                    }
                }
                if(i == count)
                    printf(" (%lu nulls)",count - 8);
                else
                {
                    /* check for all spaces                               */
                    if(fseek(inptr,offset + 8,0))
                    {
                        printf(" re-seek to comment offset %lu failed ",offset + 8);
                        why(stdout);
                    }
                    else
                    {
                        i = 8;
                        for( ; i < count; ++i)
                        {
                            if(fgetc(inptr) != ' ')
                                break;
                        }
                        if(i >= count)
                            printf(" (%d spaces)",i - 8);
                        else
                        {
                            if(fseek(inptr,offset + i,0))
                            {
                                printf(" re-seek to comment offset %lu failed ",offset + 8);
                                why(stdout);
                            }
                            else
                            {
                                (void)putchar(' ');
                                /* note that the above loop strips leading spaces */
                                if((i > 20) && (PRINT_SECTION))
                                    printf("(%d spaces)",i - 8);
                                print_ascii(inptr,count - i,offset + i);
                            }
                        }
                    }
                }
            }
            else
            {
                (void)putchar('\n');
                hexdump(inptr,offset,count,count,16,
                            0,SUBINDENT,ADDRWIDTH);
                (void)putchar('\n');
            }
        }
    }
}

#define COLORSTACKSIZE  16
static int colorstack[COLORSTACKSIZE];
static int colorstack_index = -1;

void
pushcolor(int color)
{
    if(USE_COLOR)
    {
        if(colorstack_index < 0)
        {
            colorstack_index = 0;
            colorstack[0] = END_COLOR;
        }
        else if(colorstack_index >= COLORSTACKSIZE)
            colorstack_index = 0;

        printf("\033[%dm",color);
        colorstack[++colorstack_index] = color;
    }
}

void
popcolor()
{
    if(USE_COLOR)
    {
        if(--colorstack_index < 0)
            colorstack_index = 0;
        printf("\033[%dm",colorstack[colorstack_index]);
    }
}

void
printred(char *msg)
{
    pushcolor(RED);
    printf("%s",msg);
    popcolor();
}

/* Print the Color Filter Array pattern in a human-readable form.     */
/* The array pattern is printed on a single line, as simply as        */
/* possible. This could be annoying for arrays larger than 2x2.       */

/* This prints the color characters...                                */
static void
putcolorchar(unsigned short color)
{
    switch(color)
    {
        case 0: putchar('R'); break;
        case 1: putchar('G'); break;
        case 2: putchar('B'); break;
        case 3: putchar('C'); break;
        case 4: putchar('M'); break;
        case 5: putchar('Y'); break;
        case 6: putchar('W'); break;
        default: putchar('?'); break;
    }
}

/* ...and this does the grunt work.                                   */

void
print_cfapattern(FILE *inptr,unsigned long count,unsigned short byteorder,
                unsigned long offset)
{
    unsigned short horiz_repeat;
    unsigned short vert_repeat;
    unsigned short color;
    int h,v;

    printf("\"count %lu: ",count);
    if((offset != HERE) && (fseek(inptr,offset,0) == -1))
        printf("seek failed to offset %lu to process user comment)\n",offset);
    horiz_repeat = read_ushort(inptr,byteorder,HERE);
    vert_repeat = read_ushort(inptr,byteorder,HERE);
    if(PRINT_SECTION)
    {
        printf("H=%u, V=%u ",horiz_repeat,vert_repeat);
        /* Some sanity checks                                             */
#define MAX_REPEAT  8
#define MAX_COLOR   6
        if((horiz_repeat == 0) || (vert_repeat == 0) ||
            (horiz_repeat > MAX_REPEAT) || (horiz_repeat > MAX_REPEAT))
                printf(" INVALID REPEAT COUNT");
        else
        {
            for(v = 0; v < vert_repeat; ++v)
            {
                putchar('[');
                for(h = 0; h < horiz_repeat; ++h)
                {
                    color = read_ubyte(inptr,HERE);
                    if(ferror(inptr))
                        break;
                    if(color <= MAX_COLOR)
                    {
                        putcolorchar(color);
                        putchar(',');
                    }
                    else
                    {
                        printf(" COLOR VALUE %d OUT OF RANGE",color);
                        return;
                    }
                }
                printf(" ...],");
            }
            printf(" ...\"");
        }
    }
    else
        printf("%dx%d\"",horiz_repeat,vert_repeat);
    if(ferror(inptr))
        clearerr(inptr);
}

/* Dump a bunch of bytes in traditional hex/ascii format. The offset  */
/* and size of the chunk to dump are passed, with a separate argument */
/* for the number of bytes to print, which is constrained to be <=    */
/* the size. The per-line width of the output is also passed          */
/* (sometimes this program uses 16, sometimes 12 if the chunk might   */
/* be in IFD format).                                                 */

/* This function checks to assure that the buffer in read_bytes() is  */
/* not overrun (read_bytes() can't report truncation) and loops on    */
/* the data if necessary.                                             */

/* If the length of data to be printed is less than the actual size,  */
/* the last line is annotated with "ect..." to indicate that there is */
/* more. On the other hand, if 'max_to_print' expires before the last */
/* line is full, both fields are padded with spaces.                  */

/* NOTE: this routine never prints a final newline (intentionally)    */

void
hexdump(FILE *inptr,unsigned long offset,unsigned long size,
        unsigned long max_to_print,int width,int indent,
        int subindent,int addrwidth)
{
    char *data,*next;
    int bytes = 0;
    int printed = 0;
    int etc = 0;
    unsigned long remaining,readlength;
    int i;

    if(inptr)
    {
        if((size > 0L) && (max_to_print > 0L))
        {
            if(size > MAXBUFLEN)
                readlength = MAXBUFLEN;
            else
                readlength = size;

            /* Always read in multiples of "width" if doing multiple  */
            /* reads.                                                 */
            if(size > readlength)
                readlength = (readlength / width) * width;
            if(size < max_to_print)
                max_to_print = size;

            if((offset != HERE) && (fseek(inptr,offset,0) == -1))
            {
                printf(" seek failed to offset %lu to print hex/ascii dump\n",offset);
                why(stdout);
                return;
            }
            while((printed < max_to_print) && (data = (char *)read_bytes(inptr,readlength,HERE)))
            {
                if(max_to_print < readlength)
                {
                    readlength = max_to_print;
                    ++etc;
                }

                bytes = 0;
                while(bytes < readlength)
                {
                    next = data;
                    remaining = readlength - bytes;

                    /* start address of each line                     */
                    print_tag_address(ENTRY|APP_ENTRY,offset + printed,
                                        indent + SUBINDENT," ",addrwidth);
                    extraindent(subindent); 

                    /* This is the hex part                           */
                    for(i = 0; i < width; ++i)
                    {
                        if(i && ((i % 4) == 0))
                            fputc(' ',stdout);
                        if(i < remaining)
                            fprintf(stdout,"%02x ",(unsigned int)(*data++ & 0xff));
                        else
                            printf("   ");
                    }

                    /* Now do it again for the ascii side             */
                    printf(" |");
                    remaining = readlength - bytes;
                    for(i = 0; i < width; ++i,++next,++bytes,++printed)
                    {
                        if(i >= remaining)      /* pad at the end     */
                            fputc(' ',stdout);
                        else if(isprint(*next))
                            fputc(*next,stdout);
                        else
                            fputc('.',stdout);  /* can't win 'em all  */
                    }
                    if(printed < max_to_print)
                        printf("|\n");  /* every line except the last */
                    else
                        printf("|");
                }
                if((max_to_print - printed) < readlength)
                    readlength = max_to_print - printed;
            }
            if(etc)     /* didn't print the whole chunk...            */
                printf(" etc...");
        }
    }
    else
        printf(" INTERNAL ERROR: no input file to hexdump()\n");
}

/* indent the following text by printing 'indent' spaces              */
void
putindent(int spaces)
{
    if(spaces > MAXINDENT)
        spaces = MAXINDENT;
    while(spaces-- > 0)
        (void)putchar(' ');
}

/* conditionally indent the following text                            */

void
extraindent(int spaces)
{
    if(PRINT_INDENT)
    {
        if(spaces > MAXINDENT)
            spaces = MAXINDENT;
        while(spaces-- > 0)
            (void)putchar(' ');
    }
}

#define START_VALUE " = \""
void
print_startvalue()
{
    printf("%s",START_VALUE);
}

#define END_VALUE   "\""
void
print_endvalue()
{
    printf("%s",END_VALUE);
}
